from functools import partial

import torch
from torch import Tensor, nn

from .utils import input_pre_processing

HYPERPARAMETERS = dict(
    temperature={"low": 0.1, "high": 1000, "step": 0.1}, eps={"low": 0.0, "high": 0.005, "step": 0.0001}
)


def _score_fn(x: Tensor, model: nn.Module, temperature: float = 1000, **kwargs) -> Tensor:
    outputs = model(x)
    return -(1 - torch.softmax(outputs / temperature, dim=1).square().sum(dim=1))


def doctor(x: Tensor, model: nn.Module, temperature: float = 1, eps: float = 0.0, **kwargs) -> Tensor:
    """Doctor detector.

    Args:
        x (Tensor): input tensor.
        model (nn.Module): classifier.
        temperature (float, optional): softmax temperature parameter. Defaults to 1000.
        eps (float, optional): input preprocessing noise value. Defaults to 0.0 (no input preprocessing).

    Returns:
        Tensor: scores for each input.

    References:
        [1] https://arxiv.org/abs/2106.02395
    """
    model.eval()

    if eps > 0:
        x = input_pre_processing(partial(_score_fn, model=model, temperature=temperature), x, eps)

    with torch.no_grad():
        return _score_fn(x, model, temperature)



"""Simple Tiny ImageNet dataset utility class for pytorch."""

import os
import shutil
from typing import Callable, Optional

from torchvision.datasets import ImageFolder
from torchvision.datasets.utils import download_and_extract_archive, verify_str_arg


def normalize_tin_val_folder_structure(path, images_folder="images", annotations_file="val_annotations.txt"):
    # Check if files/annotations are still there to see
    # if we already run reorganize the folder structure.
    images_folder = os.path.join(path, images_folder)
    annotations_file = os.path.join(path, annotations_file)

    # Exists
    if not os.path.exists(images_folder) and not os.path.exists(annotations_file):
        if not os.listdir(path):
            raise RuntimeError("Validation folder is empty.")
        return

    # Parse the annotations
    with open(annotations_file) as f:
        for line in f:
            values = line.split()
            img = values[0]
            label = values[1]
            img_file = os.path.join(images_folder, values[0])
            label_folder = os.path.join(path, label)
            os.makedirs(label_folder, exist_ok=True)
            try:
                shutil.move(img_file, os.path.join(label_folder, img))
            except FileNotFoundError:
                continue

    os.sync()
    assert not os.listdir(images_folder)
    shutil.rmtree(images_folder)
    os.remove(annotations_file)
    os.sync()


class TinyImageNet(ImageFolder):
    """Dataset for TinyImageNet-200"""

    base_folder = "tiny-imagenet-200"
    zip_md5 = "90528d7ca1a48142e341f4ef8d21d0de"
    splits = ("train", "val", "test")
    filename = "tiny-imagenet-200.zip"
    url = "http://cs231n.stanford.edu/tiny-imagenet-200.zip"

    def __init__(self, root, split="train", transform: Optional[Callable] = None, download=False, **kwargs):
        self.data_root = os.path.expanduser(root)
        self.split = verify_str_arg(split, "split", self.splits)

        if download:
            self.download()

        if not self._check_exists():
            raise RuntimeError("Dataset not found." + " You can use download=True to download it")
        super().__init__(self.split_folder, transform=transform, **kwargs)

    @property
    def dataset_folder(self):
        return os.path.join(self.data_root, self.base_folder)

    @property
    def split_folder(self):
        return os.path.join(self.dataset_folder, self.split)

    def _check_exists(self):
        return os.path.exists(self.split_folder)

    def extra_repr(self):
        return "Split: {split}".format(**self.__dict__)

    def download(self):
        if self._check_exists():
            return
        download_and_extract_archive(self.url, self.data_root, filename=self.filename, md5=self.zip_md5)
        assert "val" in self.splits
        normalize_tin_val_folder_structure(os.path.join(self.dataset_folder, "val"))



import torch
import torch.nn.functional as F
from torch import Tensor, nn

HYPERPARAMETERS = dict(
    temperature={"low": 0.1, "high": 1000, "step": 0.1}, eps={"low": 0.0, "high": 0.005, "step": 0.0001}
)


def odin(x: Tensor, model: nn.Module, temperature: float = 1000, eps: float = 0.0, **kwargs) -> Tensor:
    """ODIN OOD detector.

    Args:
        x (Tensor): input tensor.
        model (nn.Module): classifier.
        temperature (float, optional): softmax temperature parameter. Defaults to 1000.
        eps (float, optional): input preprocessing noise value. Defaults to 0.0 (no input preprocessing).

    Returns:
        Tensor: OOD scores for each input.

    References:
        [1] https://arxiv.org/abs/1706.02690
    """
    model.eval()
    if eps > 0:
        x.requires_grad_()
        # input preprocessing
        outputs = model(x)
        labels = torch.argmax(outputs, dim=1)
        loss = F.nll_loss(outputs / temperature, labels)
        loss.backward()

        grad_sign = x.grad.data.sign()
        # Adding small perturbations to images
        x = x - eps * grad_sign

    with torch.no_grad():
        outputs = model(x)
    return torch.softmax(outputs / temperature, dim=1).max(dim=1)[0]



import os
from typing import Callable, Optional

from torchvision.datasets import ImageFolder
from torchvision.datasets.utils import check_integrity, download_and_extract_archive


class Places365(ImageFolder):
    base_folder = "places365"
    filename = "val_256.tar"
    file_md5 = "e27b17d8d44f4af9a78502beb927f808"
    url = "http://data.csail.mit.edu/places/places365/val_256.tar"
    # size: 36500

    def __init__(
        self, root: str, split=None, transform: Optional[Callable] = None, download: bool = False, **kwargs
    ) -> None:
        self.root = os.path.expanduser(root)
        self.dataset_folder = os.path.join(self.root, self.base_folder)
        self.archive = os.path.join(self.root, self.filename)

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError("Dataset not found or corrupted." + " You can use download=True to download it")
        super().__init__(self.dataset_folder, transform=transform, **kwargs)

    def _check_integrity(self) -> bool:
        return check_integrity(self.archive, self.file_md5)

    def _check_exists(self) -> bool:
        return os.path.exists(self.dataset_folder)

    def download(self) -> None:
        if self._check_integrity() and self._check_exists():
            return
        download_and_extract_archive(
            self.url, download_root=self.root, extract_root=self.dataset_folder, md5=self.file_md5
        )



import logging
from typing import List, Literal, Optional

import torch
from torch import Tensor, nn

from detectors.methods.templates import DetectorWithFeatureExtraction

from .utils import sklearn_cov_matrix_estimarion

_logger = logging.getLogger(__name__)


def mahalanobis_dist_forward_substitution(x: Tensor, y: Tensor, L: Tensor):
    return torch.sqrt(torch.sum(torch.square(torch.mm(x, L).unsqueeze(1) - torch.mm(y, L).unsqueeze(0)), dim=-1)).min(
        dim=1, keepdim=True
    )[0]


def mahalanobis_distance_inv_fast(x: Tensor, y: Tensor, precision: Tensor):
    """Mahalanobis distance betwee x and y with an accelerated implementation.

    Args:
        x (Tensor): first point.
        y (Tensor): second point.
        precision (Tensor): inverse of the covariance matrix.
    """
    d_squared = torch.mm(torch.mm(x - y, precision), (x - y).T).diag()
    return torch.sqrt(d_squared)


def mahalanobis_distance_inv(x: Tensor, y: Tensor, precision: Tensor):
    """Mahalanobis distance betwee x and y.

    Args:
        x (Tensor): first point.
        y (Tensor): second point.
        precision (Tensor): inverse of the covariance matrix.
    """

    d_squared = torch.sum((x - y).T * torch.mm(precision, (x - y).T), dim=0)
    return torch.sqrt(d_squared)


def mahalanobis_inv_layer_score(x: Tensor, mus: Tensor, inv: Tensor) -> Tensor:
    stack = torch.zeros((x.shape[0], mus.shape[0]), device=x.device, dtype=x.dtype)
    for i, mu in enumerate(mus):
        stack[:, i] = mahalanobis_distance_inv(x, mu.reshape(1, -1), inv).reshape(-1)

    return -torch.nan_to_num(stack.min(1, keepdim=True)[0], nan=1e6)


def mahalanobis_inv_layer_score_fast(x: Tensor, mus: Tensor, inv: Tensor) -> Tensor:
    stack = torch.zeros((x.shape[0], mus.shape[0]), device=x.device, dtype=x.dtype)
    for i, mu in enumerate(mus):
        stack[:, i] = mahalanobis_distance_inv_fast(x, mu.reshape(1, -1), inv).reshape(-1)

    return -torch.nan_to_num(stack.min(1, keepdim=True)[0], nan=1e6)


def class_cond_mus_cov_inv_matrix(
    x: Tensor, targets: Tensor, cov_method: str = "EmpiricalCovariance", device=torch.device("cpu")
):
    class_cond_mean = {}
    centered_data_per_class = {}
    unique_classes = sorted(torch.unique(targets.detach().cpu()).numpy().tolist())
    for c in unique_classes:
        filt = targets == c
        temp = x[filt].to(device)
        class_cond_mean[c] = temp.mean(0, keepdim=True)
        centered_data_per_class[c] = temp - class_cond_mean[c]

        class_cond_mean[c] = class_cond_mean[c].detach().cpu()
        centered_data_per_class[c] = centered_data_per_class[c].detach().cpu()

    centered_data_per_class = torch.vstack(list(centered_data_per_class.values()))
    mus = torch.vstack(list(class_cond_mean.values()))

    mu, cov_mat, inv_mat = sklearn_cov_matrix_estimarion(centered_data_per_class.numpy(), method=cov_method)

    cov_mat = torch.from_numpy(cov_mat).float()
    inv_mat = torch.from_numpy(inv_mat).float()

    return mus, cov_mat, inv_mat


HYPERPARAMETERS = dict(
    cov_mat_method=[
        "EmpiricalCovariance",
        "GraphicalLasso",
        "GraphicalLassoCV",
        "LedoitWolf",
        "ShrunkCovariance",
        "OAS",
    ]
)


class Mahalanobis(DetectorWithFeatureExtraction):
    """Mahalanobis OOD detector.

    Args:
        model (nn.Module): Model to be used to extract features
        features_nodes (Optional[List[str]]): List of strings that represent the feature nodes.
            Defaults to None.
        all_blocks (bool, optional): If True, use all blocks of the model. Defaults to False.
        last_layer (bool, optional): If True, use also the last layer of the model. Defaults to False.
        pooling_op_name (str, optional): Pooling operation to be applied to the features.
            Can be one of `max`, `avg`, `flatten`, `getitem`, `avg_or_getitem`, `max_or_getitem`, `none`. Defaults to `avg`.
        aggregation_method_name (str, optional): Aggregation method to be applied to the features. Defaults to None.
        cov_mat_method (str, optional): Covariance matrix estimation method. Can be one of:
            `EmpiricalCovariance`, `GraphicalLasso`, `GraphicalLassoCV`, `LedoitWolf`, `MinCovDet`, `ShrunkCovariance`, `OAS`.
            Defaults to `EmpiricalCovariance`.
        mu_cov_inv_est_fn (function, optional): Function to be used to estimate the means, covariance and inverse matrix.
            Defaults to `class_cond_mus_cov_inv_matrix`.
        cov_reg (float, optional): Covariance regularization. Defaults to 1e-6.

    References:
        [1] https://arxiv.org/abs/1807.03888
    """

    def __init__(
        self,
        model: nn.Module,
        features_nodes: Optional[List[str]] = None,
        all_blocks: bool = False,
        last_layer: bool = False,
        pooling_op_name: str = "avg_or_getitem",
        aggregation_method_name: Optional[str] = "mean",
        cov_mat_method: Literal[
            "EmpiricalCovariance",
            "GraphicalLasso",
            "GraphicalLassoCV",
            "LedoitWolf",
            "MinCovDet",
            "ShrunkCovariance",
            "OAS",
        ] = "EmpiricalCovariance",
        mu_cov_inv_est_fn=class_cond_mus_cov_inv_matrix,
        cov_reg: float = 1e-6,
        **kwargs,
    ):
        super().__init__(
            model, features_nodes, all_blocks, last_layer, pooling_op_name, aggregation_method_name, **kwargs
        )
        self.cov_mat_method = cov_mat_method
        self.mu_cov_inv_est_fn = mu_cov_inv_est_fn
        self.cov_reg = cov_reg

    def _layer_score(self, x: Tensor, layer_name: Optional[str] = None, index: Optional[int] = None):
        return mahalanobis_inv_layer_score(
            x, self.mus[layer_name].to(x.device), self.precision_chols[layer_name].to(x.device)
        )

    def _fit_params(self) -> None:
        self.mus = {}
        self.invs = {}
        self.precision_chols = {}
        device = next(self.model.parameters()).device
        for layer_name, layer_features in self.train_features.items():
            self.mus[layer_name], cov, self.invs[layer_name] = self.mu_cov_inv_est_fn(
                layer_features, self.train_targets, self.cov_mat_method, device=device
            )
            cov_chol = torch.linalg.cholesky(cov.to(device) + self.cov_reg * torch.eye(cov.shape[1], device=device))
            self.precision_chols[layer_name] = torch.linalg.solve_triangular(
                cov_chol, torch.eye(cov_chol.shape[1], device=device), upper=False
            ).T



import os
from typing import Callable, Optional

import numpy as np
import torch.utils.data.dataset as dataset
from PIL import Image
from torchvision.datasets.utils import check_integrity, download_and_extract_archive, verify_str_arg

CORRUPTIONS = [
    "brightness",
    "contrast",
    "defocus_blur",
    "elastic_transform",
    "fog",
    "frost",
    "gaussian_blur",
    "gaussian_noise",
    "glass_blur",
    "impulse_noise",
    "jpeg_compression",
    "motion_blur",
    "pixelate",
    "saturate",
    "shot_noise",
    "snow",
    "spatter",
    "speckle_noise",
    "zoom_blur",
]


class CIFAR10_C(dataset.Dataset):
    base_folder = "CIFAR-10-C"
    filename = "CIFAR-10-C.tar"
    file_md5 = "56bf5dcef84df0e2308c6dcbcbbd8499"
    url = "https://zenodo.org/record/2535967/files/CIFAR-10-C.tar"
    corruptions = CORRUPTIONS

    def __init__(
        self, root: str, split: str, intensity: int, transform: Optional[Callable] = None, download: bool = False
    ) -> None:
        self.root = os.path.expanduser(root)
        self.corruption = verify_str_arg(split, "split", self.corruptions)
        self.intensity = int(intensity)
        self.transform = transform
        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError("Dataset not found or corrupted." + " You can use download=True to download it")

        indices = self._indices
        images = np.load(self._perturbation_array)[indices]
        labels = np.load(self._labels_array)[indices]
        self.arrays = [images, labels]

    def __getitem__(self, index):
        x = self.arrays[0][index]

        # to PIL image
        x = Image.fromarray(x)

        if self.transform:
            x = self.transform(x)

        y = self.arrays[1][index]
        return x, y

    def __len__(self):
        return len(self.arrays[0])

    @property
    def _indices(self):
        N = 10000
        return slice((self.intensity - 1) * N, self.intensity * N)

    @property
    def _dataset_folder(self):
        return os.path.join(self.root, self.base_folder)

    @property
    def _perturbation_array(self):
        return os.path.join(self._dataset_folder, self.corruption + ".npy")

    @property
    def _labels_array(self):
        return os.path.join(self._dataset_folder, "labels.npy")

    def _check_integrity(self) -> bool:
        root = self.root
        md5 = self.file_md5
        fpath = os.path.join(root, self.filename)
        return check_integrity(fpath, md5)

    def _check_exists(self) -> bool:
        return os.path.exists(self._perturbation_array)

    def download(self) -> None:
        if self._check_integrity() and self._check_exists():
            return
        download_and_extract_archive(self.url, download_root=self.root, md5=self.file_md5)


class CIFAR100_C(CIFAR10_C):
    base_folder = "CIFAR-100-C"
    filename = "CIFAR-100-C.tar"
    file_md5 = "11f0ed0f1191edbf9fa23466ae6021d3"
    url = "https://zenodo.org/record/3555552/files/CIFAR-100-C.tar"
    corruptions = CORRUPTIONS



from functools import partial

import torch
from torch import Tensor, nn

from detectors.methods.utils import input_pre_processing

HYPERPARAMETERS = dict(temperature=dict(low=1, high=1000, step=1.0), eps=dict(low=0, high=0.01, step=0.001))


def _score_fn(x: Tensor, model: nn.Module, temperature: float = 1.0, **kwargs) -> Tensor:
    logits = model(x)
    return temperature * torch.logsumexp(logits / temperature, dim=-1)


def energy(x: Tensor, model: nn.Module, temperature: float = 1.0, eps: float = 0.0, **kwargs):
    """Energy-based OOD detector.

    Args:
        x (Tensor): input tensor.
        model (nn.Module): classifier.
        temperature (float, optional): softmax temperature parameter. Defaults to 1.0.

    Returns:
        Tensor: OOD scores for each input.

    References:
        [1] https://arxiv.org/abs/2010.03759
    """
    model.eval()
    if eps > 0:
        x = input_pre_processing(partial(_score_fn, model=model, temperature=temperature), x, eps)

    with torch.no_grad():
        return _score_fn(x, model, temperature)



from typing import Optional, Tuple, Union


def hf_hub_url_template(model_name: str):
    return f"https://huggingface.co/edadaltocg/{model_name}/resolve/main/pytorch_model.bin"


class ModelDefaultConfig(dict):
    """
    Default configuration for models from `timm` library.

    Example::

        {
            'url': 'https://download.pytorch.org/models/resnet18-5c106cde.pth',
            'num_classes': 1000,
            'input_size': (3, 224, 224),
            'pool_size': (7, 7),
            'crop_pct': 0.875,
            'interpolation': 'bilinear',
            'fixed_input_size': True,
            'mean': (0.485, 0.456, 0.406),
            'std': (0.229, 0.224, 0.225),
            'first_conv': 'conv1',
            'classifier': 'fc',
            'architecture': 'resnet18'
        }
    """

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __init__(
        self,
        url: str,
        num_classes: int,
        input_size: Union[Tuple[int, int, int], Tuple[int, int], int],
        pool_size: Optional[Tuple[int, int]],
        crop_pct: float,
        mean: Tuple[float, float, float],
        std: Tuple[float, float, float],
        first_conv: str,
        classifier: str,
        architecture: str,
        interpolation: str = "bilinear",
        fixed_input_size: Optional[bool] = False,
        **kwargs,
    ):
        super().__init__(
            url=url,
            num_classes=num_classes,
            input_size=input_size,
            pool_size=pool_size,
            crop_pct=crop_pct,
            interpolation=interpolation,
            fixed_input_size=fixed_input_size,
            mean=mean,
            std=std,
            first_conv=first_conv,
            classifier=classifier,
            architecture=architecture,
            **kwargs,
        )



import os
from multiprocessing import cpu_count
from multiprocessing.pool import ThreadPool
from typing import Callable, Optional

from torchvision.datasets import ImageFolder
from torchvision.datasets.utils import download_url


class OpenImageO(ImageFolder):
    """OpenImageO dataset.

    - Length: <17632
        (sore url returns http error code 404.
        Might need to invoke several times to get at lkeas 16k images). Confirmed 16067 images.
    - Size: >30GB
    - Paper: https://arxiv.org/pdf/2203.10807.pdf
    - Auxiliary file: `OpenImageO/openimage_o_urls.csv`

    Args:
    """

    base_folder = "openimage-o"
    filenames = "OpenImageO/openimage_o_urls.csv"
    # size: 17632

    def __init__(
        self, root: str, split=None, transform: Optional[Callable] = None, download: bool = False, **kwargs
    ) -> None:
        self.root = os.path.expanduser(root)
        self.dataset_folder = os.path.join(self.root, self.base_folder)
        self.id_urls = [
            l.split(",") for l in open(os.path.join(os.path.dirname(__file__), self.filenames)).read().splitlines()[1:]
        ]

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError("Dataset not found or corrupted. You can use download=True to download it")

        super().__init__(self.dataset_folder, transform=transform, **kwargs)

    def _check_integrity(self) -> bool:
        # assert number of iumages in folder is equal to 17632
        if not self._check_exists():
            return False

        count = 0
        # Iterate directory
        for path in os.listdir(os.path.join(self.dataset_folder, self.base_folder)):
            # check if current path is a file
            if os.path.isfile(os.path.join(self.dataset_folder, self.base_folder, path)):
                count += 1
        return count >= 16_000

    def _check_exists(self) -> bool:
        return os.path.exists(self.dataset_folder)

    def download(self) -> None:
        if self._check_integrity() and self._check_exists():
            return

        def download_url_wrapper(args):
            url, id, root = args
            extension = url.split(".")[-1]
            filename = id + "." + extension
            if not os.path.exists(os.path.join(root, filename)):
                try:
                    download_url(url, root, filename=filename, max_redirect_hops=10)
                except Exception as e:
                    print(f"Error downloading: {url}. Error: {e}")

        args = [(url, id, os.path.join(self.dataset_folder, self.base_folder)) for id, url in self.id_urls]
        cpus = cpu_count()
        with ThreadPool(cpus - 1) as pool:
            results = pool.map_async(download_url_wrapper, args)
            results.wait()

            pool.close()
            pool.join()



import torch
from torch import Tensor, nn


def entropy(x: Tensor, model: nn.Module, **kwargs) -> Tensor:
    with torch.no_grad():
        logits = model(x)
    return torch.sum(-torch.softmax(logits, dim=1) * torch.log_softmax(logits, dim=1), dim=1)



"""
Configuration for the project.

It is used to set the default paths for the data, checkpoints, results, etc.

Constants:

- `DATA_DIR`: The directory where the data is stored.
- `IMAGENET_ROOT`: The directory where the ImageNet data is stored.
- `CHECKPOINTS_DIR`: The directory where the checkpoints are stored.
- `RESULTS_DIR`: The directory where the results are stored.
"""
import os

HOME = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
DATA_DIR = os.environ.get("DATA_DIR", os.path.join(HOME, "data/"))
IMAGENET_ROOT = os.environ.get("IMAGENET_ROOT", DATA_DIR)
CHECKPOINTS_DIR = os.environ.get("CHECKPOINTS_DIR", os.path.join(HOME, "checkpoints/"))
RESULTS_DIR = os.environ.get("RESULTS_DIR", os.path.join(HOME, "results/"))



"""Finetuned ViT models for CIFAR10, CIFAR100, and SVHN datasets."""
import timm
import timm.models
import torch
import torch.nn as nn
from timm.models import register_model as timm_register_model

from detectors.models.utils import ModelDefaultConfig, hf_hub_url_template


def _cfg(url, architecture: str, **kwargs):
    model = timm.create_model(architecture, pretrained=False)
    base_cfg = model.default_cfg
    num_classes = kwargs.pop("num_classes", 10)
    # replace base cfg with new arguments
    base_cfg.update(kwargs)
    base_cfg["architecture"] = architecture
    base_cfg["url"] = url
    base_cfg["num_classes"] = num_classes
    return ModelDefaultConfig(**base_cfg)


default_cfgs = {
    "vit_base_patch16_224_in21k_ft_cifar10": _cfg(
        url=hf_hub_url_template("vit_base_patch16_224_in21k_ft_cifar10"),
        architecture="timm/vit_base_patch16_224.orig_in21k_ft_in1k",
    ),
    "vit_base_patch16_224_in21k_ft_cifar100": _cfg(
        url=hf_hub_url_template("vit_base_patch16_224_in21k_ft_cifar100"),
        num_classes=100,
        architecture="timm/vit_base_patch16_224.orig_in21k_ft_in1k",
    ),
    "vit_base_patch16_224_in21k_ft_svhn": _cfg(
        url=hf_hub_url_template("vit_base_patch16_224_in21k_ft_svhn"),
        architecture="timm/vit_base_patch16_224.orig_in21k_ft_in1k",
    ),
}


def _create_vit_ft(variant, pretrained=False, **kwargs):
    default_cfg = default_cfgs[variant]

    # load timm model
    model = timm.create_model(default_cfg["architecture"], pretrained=not pretrained)
    # override timm config
    model.default_cfg = default_cfg

    # override model
    model.head = nn.Linear(model.head.in_features, default_cfg["num_classes"])

    if pretrained:
        model.load_state_dict(
            torch.hub.load_state_dict_from_url(default_cfg["url"], map_location="cpu", file_name=f"{variant}.pth")
        )

    return model


@timm_register_model
def vit_base_patch16_224_in21k_ft_cifar10(pretrained=False, **kwargs):
    return _create_vit_ft("vit_base_patch16_224_in21k_ft_cifar10", pretrained=pretrained, **kwargs)


@timm_register_model
def vit_base_patch16_224_in21k_ft_cifar100(pretrained=False, **kwargs):
    return _create_vit_ft("vit_base_patch16_224_in21k_ft_cifar100", pretrained=pretrained, **kwargs)


@timm_register_model
def vit_base_patch16_224_in21k_ft_svhn(pretrained=False, **kwargs):
    return _create_vit_ft("vit_base_patch16_224_in21k_ft_svhn", pretrained=pretrained, **kwargs)



"""From https://github.com/KaihuaTang/Long-Tailed-Recognition.pytorch/blob/master/classification/data/dataloader.py"""
import os

import numpy as np
from PIL import Image
from torch.utils.data import Dataset


class LT_Dataset(Dataset):
    def __init__(self, root: str, txt_path: str, subset_class_idx=np.arange(0, 1000), transform=None):
        self.img_path = []
        self.labels = []
        self.transform = transform
        with open(txt_path) as f:
            for line in f:
                _label = int(line.split()[1])
                if _label in subset_class_idx:
                    self.labels.append(_label)
                    self.img_path.append(os.path.join(root, line.split()[0]))

        num_classes = len(subset_class_idx)
        self.img_num_per_cls = np.zeros(num_classes)
        for i in range(num_classes):
            self.img_num_per_cls[i] = np.sum(np.array(self.labels) == subset_class_idx[i])

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, index):

        path = self.img_path[index]
        label = self.labels[index]

        with open(path, "rb") as f:
            sample = Image.open(f).convert("RGB")

        if self.transform is not None:
            sample = self.transform(sample)

        return sample, label



import logging
from functools import partial
from typing import Optional

import torch
import torch.nn.functional as F
from torch import Tensor

from detectors.methods.utils import input_pre_processing

_logger = logging.getLogger(__name__)


HYPERPARAMETERS = dict(temperature=dict(low=0.1, high=1000, step=0.1), eps=dict(low=0.0, high=0.005, step=0.0001))


def igeoodlogits_vec(logits, temperature, centroids, epsilon=1e-12):
    logits = torch.sqrt(F.softmax(logits / temperature, dim=1))
    centroids = torch.sqrt(F.softmax(centroids / temperature, dim=1))
    mult = logits @ centroids.T
    stack = 2 * torch.acos(torch.clamp(mult, -1 + epsilon, 1 - epsilon))
    return stack


def _score_fn(x: Tensor, model: torch.nn.Module, centroids: Tensor, temperature: float = 1.0, **kwargs) -> Tensor:
    logits = model(x)
    return igeoodlogits_vec(logits, temperature, centroids).mean(dim=1)


class IgeoodLogits:
    """IGEOOD detector.

    Args:
        model (nn.Module): classifier.
        temperature (float, optional): softmax temperature parameter. Defaults to 1.0.
        eps (float, optional): input preprocessing noise value. Defaults to 0.0 (no input preprocessing).

    References:
        [1] https://arxiv.org/abs/2203.07798
    """

    def __init__(self, model: torch.nn.Module, temperature: float = 1.0, eps: float = 0.0, **kwargs):
        self.model = model
        self.temperature = temperature
        self.eps = eps

        self.model.eval()

    @torch.no_grad()
    def start(self, example: Optional[Tensor] = None, fit_length: Optional[int] = None, *args, **kwargs):
        self.train_features = []
        self.train_targets = []
        self.mus = []
        self.idx = 0
        if example is not None and fit_length is not None:
            logits = self.model(example)
            self.train_features = torch.zeros((fit_length,) + logits.shape[1:], dtype=logits.dtype)
            self.train_targets = torch.ones((fit_length,), dtype=torch.long) * -1

    @torch.no_grad()
    def update(self, x: Tensor, y: Tensor, *args, **kwargs):
        self.batch_size = x.shape[0]
        logits = self.model(x)

        if isinstance(self.train_features, list):
            self.train_features.append(logits)
        else:
            self.train_features[self.idx : self.idx + logits.shape[0]] = logits

        if isinstance(self.train_targets, list):
            self.train_targets.append(y)
        else:
            self.train_targets[self.idx : self.idx + y.shape[0]] = y

        self.idx += y.shape[0]

    def end(self, *args, **kwargs):
        if isinstance(self.train_features, list):
            self.train_features = torch.cat(self.train_features, dim=0)
        else:
            self.train_features = self.train_features[: self.idx]
        if isinstance(self.train_targets, list):
            self.train_targets = torch.cat(self.train_targets, dim=0)
        else:
            self.train_targets = self.train_targets[: self.idx]
        assert torch.all(self.train_targets > -1), "Not all targets were updated"

        self._fit_params()

        del self.train_features
        del self.train_targets

    def _fit_params(self) -> None:
        self.mus = []
        unique_classes = torch.unique(self.train_targets).detach().cpu().numpy().tolist()
        for c in unique_classes:
            filt = self.train_targets == c
            if filt.sum() == 0:
                continue
            self.mus.append(self.train_features[filt].mean(0, keepdim=True))
        self.mus = torch.cat(self.mus, dim=0)

    def __call__(self, x: Tensor) -> Tensor:
        self.mus = self.mus.to(x.device)
        if self.eps > 0:
            x = input_pre_processing(
                partial(_score_fn, model=self.model, temperature=self.temperature, centroids=self.mus), x, self.eps
            )

        with torch.no_grad():
            return _score_fn(x, self.model, self.mus, temperature=self.temperature)



import logging
from typing import Optional

import torch
from torch import Tensor

from detectors.methods.knn_euclides import KnnEuclides

_logger = logging.getLogger(__name__)

HYPERPARAMETERS = dict(k=dict(low=1, high=100, step=5), avg_topk=[True, False])


class KnnProjection(KnnEuclides):
    """K-NN detector based on perpendicular vector projection.

    Args:
        model (nn.Module): Model to be used to extract features
        features_nodes (Optional[List[str]]): List of strings that represent the feature nodes.
            Defaults to None.
        all_blocks (bool, optional): If True, use all blocks of the model. Defaults to False.
        last_layer (bool, optional): If True, use also the last layer of the model. Defaults to False.
        pooling_op_name (str, optional): Pooling operation to be applied to the features.
            Can be one of ["max", "avg", "flatten", "getitem", "none"]. Defaults to "avg".
        aggregation_method_name (str, optional): Aggregation method to be applied to the features. Defaults to None.
        alpha (float, optional): Alpha parameter for the input pre-processing. Defaults to 1.
        k (int, optional): Number of nearest neighbors to be considered. Defaults to 10.
        avg_topk (bool, optional): If True, average the top-k scores. Defaults to False.
    """

    def _layer_score(self, x: Tensor, layer_name: Optional[str] = None, index: Optional[int] = None):
        pairwise = x @ self.ref[layer_name].to(x.device).T
        topk, _ = torch.topk(pairwise, k=self.k, dim=-1)
        if self.mean_op:
            return topk.mean(dim=-1)
        else:
            return topk[:, -1]



from typing import Optional

import torch
from torch import Tensor, nn
from torchvision.models.feature_extraction import create_feature_extractor

HYPERPARAMETERS = dict(temperature=dict(low=0.1, high=1000, step=0.1))


def gradnorm(x: Tensor, model: nn.Module, last_layer_name: Optional[str] = None, temperature: float = 1.0, **kwargs):
    """GradNorm OOD detector.

    Args:
        x (Tensor): input tensor.
        model (nn.Module): classifier.
        last_layer_name (Optional[str], optional): last layer node name. Defaults to None.
            If None, the last layer is automatically selected.
        temperature (float, optional): softmax temperature parameter. Defaults to 1.0.

    Returns:
        Tensor: scores for each input.

    References:
        [1] https://arxiv.org/abs/2110.00218
    """

    if last_layer_name is None:
        last_layer_name = list(model._modules.keys())[-1]
    last_layer = model._modules[last_layer_name]
    assert isinstance(last_layer, nn.Linear), "Last layer must be a linear layer"
    # feature extractor
    penultimate_layer_name = list(model._modules.keys())[-2]
    feature_extractor = create_feature_extractor(model, [penultimate_layer_name])
    with torch.no_grad():
        features = feature_extractor(x)[penultimate_layer_name]

    scores = torch.empty(x.shape[0], dtype=torch.float32, device=x.device)
    for i, l in enumerate(features):
        l = l.unsqueeze(0)
        last_layer.zero_grad()
        loss = torch.mean(torch.sum(-torch.log_softmax(last_layer(l) / temperature, dim=-1), dim=-1))
        loss.backward()
        layer_grad_norm = torch.sum(torch.abs(last_layer.weight.grad.data))
        scores[i] = layer_grad_norm
    return scores



"""Base abstract pipeline class."""
import logging
from typing import Any, Dict

logger = logging.getLogger(__name__)


class Pipeline:
    """Base Pipeline class."""

    def __init__(self) -> None:
        self.setup()

    def save_pretrained(self, path):
        raise NotImplementedError

    def load_pretrained(self, path):
        raise NotImplementedError

    def setup(self, *args, **kwargs):
        return

    def preprocess(self, *args, **kwargs) -> Any:
        return

    def postprocess(self, *args, **kwargs):
        return

    def run(self, *args, **kwargs) -> Dict[str, Any]:
        raise NotImplementedError

    def report(self, *args, **kwargs) -> str:
        raise NotImplementedError



import logging
from inspect import isclass

import numpy as np
import torch
from torch import linalg
from torch.distributions.multivariate_normal import MultivariateNormal
from tqdm import tqdm

_logger = logging.getLogger(__name__)


def check_is_fitted(estimator, attributes=None, *, msg=None, all_or_any=all):
    """Perform is_fitted validation for estimator.

    Checks if the estimator is fitted by verifying the presence of
    fitted attributes (ending with a trailing underscore) and otherwise
    raises a NotFittedError with the given message.

    If an estimator does not set any attributes with a trailing underscore, it
    can define a ``__sklearn_is_fitted__`` method returning a boolean to specify if the
    estimator is fitted or not.

    Parameters
    ----------
    estimator : estimator instance
        Estimator instance for which the check is performed.

    attributes : str, list or tuple of str, default=None
        Attribute name(s) given as string or a list/tuple of strings
        Eg.: ``["coef_", "estimator_", ...], "coef_"``

        If `None`, `estimator` is considered fitted if there exist an
        attribute that ends with a underscore and does not start with double
        underscore.

    msg : str, default=None
        The default error message is, "This %(name)s instance is not fitted
        yet. Call 'fit' with appropriate arguments before using this
        estimator."

        For custom messages if "%(name)s" is present in the message string,
        it is substituted for the estimator name.

        Eg. : "Estimator, %(name)s, must be fitted before sparsifying".

    all_or_any : callable, {all, any}, default=all
        Specify whether all or any of the given attributes must exist.

    Raises
    ------
    TypeError
        If the estimator is a class or not an estimator instance

    NotFittedError
        If the attributes are not found.
    """
    if isclass(estimator):
        raise TypeError("{} is a class, not an instance.".format(estimator))
    if msg is None:
        msg = (
            "This %(name)s instance is not fitted yet. Call 'fit' with "
            "appropriate arguments before using this estimator."
        )

    if not hasattr(estimator, "fit"):
        raise TypeError("%s is not an estimator instance." % (estimator))

    if attributes is not None:
        if not isinstance(attributes, (list, tuple)):
            attributes = [attributes]
        fitted = all_or_any([hasattr(estimator, attr) for attr in attributes])
    elif hasattr(estimator, "__sklearn_is_fitted__"):
        fitted = estimator.__sklearn_is_fitted__()
    else:
        fitted = [v for v in vars(estimator) if v.endswith("_") and not v.startswith("__")]

    if not fitted:
        raise


def _estimate_gaussian_covariances_full(resp, X, nk, means, reg_covar):
    """Estimate the full covariance matrices.

    Parameters
    ----------
    resp : array-like of shape (n_samples, n_components)

    X : array-like of shape (n_samples, n_features)

    nk : array-like of shape (n_components,)

    means : array-like of shape (n_components, n_features)

    reg_covar : float

    Returns
    -------
    covariances : array, shape (n_components, n_features, n_features)
        The covariance matrix of the current components.
    """
    n_components, n_features = means.shape
    covariances = torch.empty((n_components, n_features, n_features), dtype=X.dtype)
    for k in range(n_components):
        diff = X - means[k]
        cov = torch.mm(resp[:, k] * diff.T, diff) / nk[k]
        trace = cov.diag().sum()
        covariances[k] = cov.cpu()
        covariances[k].flatten()[:: n_features + 1] += reg_covar * trace.cpu()  # add reg_covar to the diagonal
    return covariances


def _estimate_gaussian_covariances_tied(resp, X, nk, means, reg_covar):
    """Estimate the tied covariance matrix.

    Parameters
    ----------
    resp : array-like of shape (n_samples, n_components)

    X : array-like of shape (n_samples, n_features)

    nk : array-like of shape (n_components,)

    means : array-like of shape (n_components, n_features)

    reg_covar : float

    Returns
    -------
    covariance : array, shape (n_features, n_features)
        The tied covariance matrix of the components.
    """
    avg_X2 = torch.mm(X.T, X)
    avg_means2 = torch.mm(nk * means.T, means)
    covariance = avg_X2 - avg_means2
    covariance /= nk.sum()
    trace = covariance.diag().sum()
    covariance.flatten()[:: len(covariance) + 1] += reg_covar * trace
    return covariance


def _estimate_gaussian_covariances_diag(resp, X, nk, means, reg_covar):
    """Estimate the diagonal covariance vectors.

    Parameters
    ----------
    responsibilities : array-like of shape (n_samples, n_components)

    X : array-like of shape (n_samples, n_features)

    nk : array-like of shape (n_components,)

    means : array-like of shape (n_components, n_features)

    reg_covar : float

    Returns
    -------
    covariances : array, shape (n_components, n_features)
        The covariance vector of the current components.
    """
    avg_X2 = torch.mm(resp.T, X * X) / nk[:, np.newaxis]
    avg_means2 = means**2
    avg_X_means = means * torch.mm(resp.T, X) / nk[:, np.newaxis]
    return avg_X2 - 2 * avg_X_means + avg_means2 + reg_covar


def _estimate_gaussian_covariances_spherical(resp, X, nk, means, reg_covar):
    """Estimate the spherical variance values.

    Parameters
    ----------
    responsibilities : array-like of shape (n_samples, n_components)

    X : array-like of shape (n_samples, n_features)

    nk : array-like of shape (n_components,)

    means : array-like of shape (n_components, n_features)

    reg_covar : float

    Returns
    -------
    variances : array, shape (n_components,)
        The variance values of each components.
    """
    return _estimate_gaussian_covariances_diag(resp, X, nk, means, reg_covar).mean(1)


def _estimate_gaussian_parameters(X, resp, reg_covar, covariance_type):
    """Estimate the Gaussian distribution parameters.

    Parameters
    ----------
    X : array-like of shape (n_samples, n_features)
        The input data array.

    resp : array-like of shape (n_samples, n_components)
        The responsibilities for each data sample in X.

    reg_covar : float
        The regularization added to the diagonal of the covariance matrices.

    covariance_type : {'full', 'tied', 'diag'}
        The type of precision matrices.

    Returns
    -------
    nk : array-like of shape (n_components,)
        The numbers of data samples in the current components.

    means : array-like of shape (n_components, n_features)
        The centers of the current components.

    covariances : array-like
        The covariance matrix of the current components.
        The shape depends of the covariance_type.
    """
    nk = resp.sum(dim=0) + 10 * torch.finfo(resp.dtype).eps
    means = resp.T @ X / nk[:, np.newaxis]
    covariances = {
        "full": _estimate_gaussian_covariances_full,
        "tied": _estimate_gaussian_covariances_tied,
        "diag": _estimate_gaussian_covariances_diag,
        # "spherical": _estimate_gaussian_covariances_spherical,
    }[covariance_type](resp, X, nk, means, reg_covar)
    return nk, means, covariances


def _compute_precision_cholesky(covariances, covariance_type, device="cpu"):
    """Compute the Cholesky decomposition of the precisions.

    Parameters
    ----------
    covariances : array-like
        The covariance matrix of the current components.
        The shape depends of the covariance_type.

    covariance_type : {'full', 'tied', 'diag'}
        The type of precision matrices.

    Returns
    -------
    precisions_cholesky : array-like
        The cholesky decomposition of sample precisions of the current
        components. The shape depends of the covariance_type.
    """

    if covariance_type == "full":
        n_components, n_features, _ = covariances.shape
        precisions_chol = torch.empty((n_components, n_features, n_features), dtype=covariances.dtype)
        for k, covariance in enumerate(covariances):
            cov_chol = linalg.cholesky(covariance.to(device), upper=False)
            precisions_chol[k] = linalg.solve_triangular(
                cov_chol, torch.eye(n_features, device=cov_chol.device), upper=False
            ).T.cpu()
    elif covariance_type == "tied":
        _, n_features = covariances.shape
        cov_chol = linalg.cholesky(covariances, upper=False)
        precisions_chol = linalg.solve_triangular(
            cov_chol, torch.eye(n_features, device=cov_chol.device), upper=False
        ).T
    else:
        if torch.any(torch.less_equal(covariances, 0.0)):
            raise ValueError
        precisions_chol = 1.0 / torch.sqrt(covariances)
    return precisions_chol


###############################################################################
# Gaussian mixture probability estimators
def _compute_log_det_cholesky(matrix_chol, covariance_type, n_features):
    """Compute the log-det of the cholesky decomposition of matrices.

    Parameters
    ----------
    matrix_chol : array-like
        Cholesky decompositions of the matrices.
        'full' : shape of (n_components, n_features, n_features)
        'tied' : shape of (n_features, n_features)
        'diag' : shape of (n_components, n_features)

    covariance_type : {'full', 'tied', 'diag'}

    n_features : int
        Number of features.

    Returns
    -------
    log_det_precision_chol : array-like of shape (n_components,)
        The determinant of the precision matrix for each component.
    """
    if covariance_type == "full":
        n_components, _, _ = matrix_chol.shape
        log_det_chol = torch.sum(torch.log(matrix_chol.reshape(n_components, -1)[:, :: n_features + 1]), 1)

    elif covariance_type == "tied":
        log_det_chol = torch.sum(torch.log(torch.diag(matrix_chol)))

    elif covariance_type == "diag":
        log_det_chol = torch.sum(torch.log(matrix_chol), dim=1)

    else:
        log_det_chol = n_features * (torch.log(matrix_chol))

    return log_det_chol


def _estimate_log_gaussian_prob(X, means, precisions_chol, covariance_type):
    """Estimate the log Gaussian probability.

    Parameters
    ----------
    X : array-like of shape (n_samples, n_features)

    means : array-like of shape (n_components, n_features)

    precisions_chol : array-like
        Cholesky decompositions of the precision matrices.
        'full' : shape of (n_components, n_features, n_features)
        'tied' : shape of (n_features, n_features)
        'diag' : shape of (n_components, n_features)

    covariance_type : {'full', 'tied', 'diag'}

    Returns
    -------
    log_prob : array, shape (n_samples, n_components)
    """
    n_samples, n_features = X.shape
    n_components, _ = means.shape
    # The determinant of the precision matrix from the Cholesky decomposition
    # corresponds to the negative half of the determinant of the full precision
    # matrix.
    # In short: det(precision_chol) = - det(precision) / 2
    log_det = _compute_log_det_cholesky(precisions_chol.to(X.device), covariance_type, n_features)

    if covariance_type == "full":
        log_prob = torch.empty((n_samples, n_components), dtype=X.dtype)
        for k, (mu, prec_chol) in enumerate(zip(means, precisions_chol)):
            y = torch.mm(X, prec_chol.to(X.device)) - torch.mm(mu.unsqueeze(0).to(X.device), prec_chol.to(X.device))
            log_prob[:, k] = torch.sum(torch.square(y), dim=1).cpu()

    elif covariance_type == "tied":
        log_prob = torch.empty((n_samples, n_components), dtype=X.dtype)
        for k, mu in enumerate(means):
            y = torch.mm(X, precisions_chol.to(X.device)) - torch.mm(
                mu.unsqueeze(0).to(X.device), precisions_chol.to(X.device)
            )
            log_prob[:, k] = torch.sum(torch.square(y), dim=1).cpu()

    elif covariance_type == "diag":
        precisions = precisions_chol**2
        log_prob = (
            torch.sum((means.to(X.device) ** 2 * precisions.to(X.device)), 1)
            - 2.0 * torch.mm(X, (means.to(X.device) * precisions.to(X.device)).T)
            + torch.mm(X**2, precisions.to(X.device).T)
        )

    elif covariance_type == "spherical":
        raise NotImplementedError("spherical covariance is not implemented yet")
        # precisions = precisions_chol**2
        # log_prob = (
        #     torch.sum(means**2, 1) * precisions
        #     - 2 * torch.mm(X, means.T * precisions)
        #     + torch.outer(row_norms(X, squared=True), precisions)
        # )
    # Since we are using the precision of the Cholesky decomposition,
    # `- 0.5 * log_det_precision` becomes `+ log_det_precision_chol`
    return -0.5 * (n_features * torch.log(2 * torch.tensor(torch.pi)) + log_prob) + log_det.to(log_prob.device)


class GaussianMixture:
    def __init__(
        self,
        n_components=1,
        *,
        covariance_type="full",
        tol=1e-3,
        reg_covar=1e-4,
        max_iter=100,
        n_init=1,
        init_params="kmeans",
        weights_init=None,
        means_init=None,
        precisions_init=None,
        random_state=None,
        warm_start=False,
        verbose=0,
        verbose_interval=10,
    ):
        self.n_components = n_components
        self.tol = tol
        self.reg_covar = reg_covar
        self.max_iter = max_iter
        self.n_init = n_init
        self.init_params = init_params
        self.random_state = random_state
        self.warm_start = warm_start
        self.verbose = verbose
        self.verbose_interval = verbose_interval
        self.covariance_type = covariance_type
        self.weights_init = weights_init
        self.means_init = means_init
        self.precisions_init = precisions_init
        self.random_state = random_state

    def _check_parameters(self, X):
        """Check the Gaussian mixture parameters are well defined."""
        _, n_features = X.shape

    def _initialize(self, X, resp):
        """Initialization of the Gaussian mixture parameters.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)

        resp : array-like of shape (n_samples, n_components)
        """
        n_samples, _ = X.shape

        weights, means, covariances = _estimate_gaussian_parameters(X, resp, self.reg_covar, self.covariance_type)
        weights /= n_samples

        self.weights_ = weights if self.weights_init is None else self.weights_init
        self.means_ = means if self.means_init is None else self.means_init

        if self.precisions_init is None:
            self.covariances_ = covariances
            self.precisions_cholesky_ = _compute_precision_cholesky(covariances, self.covariance_type, device=X.device)
        elif self.covariance_type == "full":
            self.precisions_cholesky_ = torch.tensor(
                [linalg.cholesky(prec_init, upper=False) for prec_init in self.precisions_init]
            )
        elif self.covariance_type == "tied":
            self.precisions_cholesky_ = linalg.cholesky(self.precisions_init, upper=False)
        else:
            self.precisions_cholesky_ = torch.sqrt(self.precisions_init)

    def _m_step(self, X, log_resp):
        """M step.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)

        log_resp : array-like of shape (n_samples, n_components)
            Logarithm of the posterior probabilities (or responsibilities) of
            the point of each sample in X.
        """
        self.weights_, self.means_, self.covariances_ = _estimate_gaussian_parameters(
            X, torch.exp(log_resp), self.reg_covar, self.covariance_type
        )
        self.weights_ /= self.weights_.sum()
        self.precisions_cholesky_ = _compute_precision_cholesky(
            self.covariances_, self.covariance_type, device=X.device
        )

    def _estimate_log_prob(self, X):
        return _estimate_log_gaussian_prob(X, self.means_, self.precisions_cholesky_, self.covariance_type).to(X.device)

    def _estimate_log_weights(self):
        return torch.log(self.weights_)

    def _compute_lower_bound(self, _, log_prob_norm):
        return log_prob_norm

    def _get_parameters(self):
        return (
            self.weights_,
            self.means_,
            self.covariances_,
            self.precisions_cholesky_,
        )

    def _set_parameters(self, params):
        (
            self.weights_,
            self.means_,
            self.covariances_,
            self.precisions_cholesky_,
        ) = params

        # Attributes computation
        _, n_features = self.means_.shape

        if self.covariance_type == "full":
            self.precisions_ = torch.empty(self.precisions_cholesky_.shape)
            for k, prec_chol in enumerate(self.precisions_cholesky_):
                self.precisions_[k] = torch.mm(prec_chol, prec_chol.T)

        elif self.covariance_type == "tied":
            self.precisions_ = torch.mm(self.precisions_cholesky_, self.precisions_cholesky_.T)
        else:
            self.precisions_ = self.precisions_cholesky_**2

    def _n_parameters(self):
        """Return the number of free parameters in the model."""
        _, n_features = self.means_.shape
        if self.covariance_type == "full":
            cov_params = self.n_components * n_features * (n_features + 1) / 2.0
        elif self.covariance_type == "diag":
            cov_params = self.n_components * n_features
        elif self.covariance_type == "tied":
            cov_params = n_features * (n_features + 1) / 2.0
        elif self.covariance_type == "spherical":
            cov_params = self.n_components
        mean_params = n_features * self.n_components
        return int(cov_params + mean_params + self.n_components - 1)

    def _initialize_parameters(self, X):
        """Initialize the model parameters.

        Parameters
        ----------
        X : array-like of shape  (n_samples, n_features)

        random_state : RandomState
            A random number generator instance that controls the random seed
            used for the method chosen to initialize the parameters.
        """
        n_samples, _ = X.shape

        if self.init_params == "kmeans":
            raise NotImplementedError("Kmeans initialization is not implemented for GaussianMixture")
            # resp = torch.zeros((n_samples, self.n_components))
            # label = (
            #     cluster.KMeans(n_clusters=self.n_components, n_init=1, random_state=self.random_state).fit(X).labels_
            # )
            # resp[torch.arange(n_samples), label] = 1
        elif self.init_params == "random":
            resp = torch.rand(size=(n_samples, self.n_components), device=X.device, dtype=X.dtype)
            resp /= resp.sum(dim=1)[:, np.newaxis]
        elif self.init_params == "random_from_data":
            resp = torch.zeros((n_samples, self.n_components), device=X.device, dtype=X.dtype)
            indices = torch.randperm(n_samples)
            resp[indices, :] = 1
        elif self.init_params == "k-means++":
            raise NotImplementedError("Kmeans++ initialization is not implemented for GaussianMixture")
            # resp = torch.zeros((n_samples, self.n_components))
            # _, indices = kmeans_plusplus(
            #     X,
            #     self.n_components,
            #     random_state=self.random_state,
            # )
            # resp[indices, torch.arange(self.n_components)] = 1
        else:
            raise ValueError("Unimplemented initialization method '%s'" % self.init_params)

        self._initialize(X, resp)

    def bic(self, X):
        """Bayesian information criterion for the current model on the input X.

        Parameters
        ----------
        X : array of shape (n_samples, n_dimensions)
            The input samples.

        Returns
        -------
        bic : float
            The lower the better.
        """
        return -2 * self.score(X) * X.shape[0] + self._n_parameters() * torch.log(X.shape[0])

    def aic(self, X):
        """Akaike information criterion for the current model on the input X.

        Parameters
        ----------
        X : array of shape (n_samples, n_dimensions)
            The input samples.

        Returns
        -------
        aic : float
            The lower the better.
        """
        return -2 * self.score(X) * X.shape[0] + 2 * self._n_parameters()

    def fit(self, X, y=None):
        """Estimate model parameters with the EM algorithm.

        The method fits the model ``n_init`` times and sets the parameters with
        which the model has the largest likelihood or lower bound. Within each
        trial, the method iterates between E-step and M-step for ``max_iter``
        times until the change of likelihood or lower bound is less than
        ``tol``, otherwise, a ``ConvergenceWarning`` is raised.
        If ``warm_start`` is ``True``, then ``n_init`` is ignored and a single
        initialization is performed upon the first call. Upon consecutive
        calls, training starts where it left off.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : object
            The fitted mixture.
        """
        # parameters are validated in fit_predict
        self.fit_predict(X, y)
        return self

    def fit_predict(self, X, y=None):
        """Estimate model parameters using X and predict the labels for X.

        The method fits the model n_init times and sets the parameters with
        which the model has the largest likelihood or lower bound. Within each
        trial, the method iterates between E-step and M-step for `max_iter`
        times until the change of likelihood or lower bound is less than
        `tol`, otherwise, a :class:`~sklearn.exceptions.ConvergenceWarning` is
        raised. After fitting, it predicts the most probable label for the
        input data points.

        .. versionadded:: 0.20

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        labels : array, shape (n_samples,)
            Component labels.
        """
        if X.shape[0] < self.n_components:
            raise ValueError(
                "Expected n_samples >= n_components "
                f"but got n_components = {self.n_components}, "
                f"n_samples = {X.shape[0]}"
            )
        self._check_parameters(X)

        # if we enable warm_start, we will have a unique initialisation
        do_init = not (self.warm_start and hasattr(self, "converged_"))
        n_init = self.n_init if do_init else 1

        max_lower_bound = -torch.inf
        self.converged_ = False

        n_samples, _ = X.shape
        for init in range(n_init):
            if do_init:
                self._initialize_parameters(X)

            lower_bound = -torch.inf if do_init else self.lower_bound_

            if self.max_iter == 0:
                best_params = self._get_parameters()
                best_n_iter = 0
            else:
                for n_iter in tqdm(range(1, self.max_iter + 1), desc="EM iteration"):
                    prev_lower_bound = lower_bound

                    log_prob_norm, log_resp = self._e_step(X)
                    self._m_step(X, log_resp)
                    lower_bound = self._compute_lower_bound(log_resp, log_prob_norm)

                    change = lower_bound - prev_lower_bound

                    if abs(change) < self.tol:
                        self.converged_ = True
                        break

                if lower_bound > max_lower_bound or max_lower_bound == -torch.inf:
                    max_lower_bound = lower_bound
                    best_params = self._get_parameters()
                    best_n_iter = n_iter

        # Should only warn about convergence if max_iter > 0, otherwise
        # the user is assumed to have used 0-iters initialization
        # to get the initial means.
        if not self.converged_ and self.max_iter > 0:
            _logger.warn(
                "Initialization %d did not converge. "
                "Try different init parameters, "
                "or increase max_iter, tol "
                "or check for degenerate data." % (init + 1),
            )

        self._set_parameters(best_params)
        self.n_iter_ = best_n_iter
        self.lower_bound_ = max_lower_bound

        # Always do a final e-step to guarantee that the labels returned by
        # fit_predict(X) are always consistent with fit(X).predict(X)
        # for any value of max_iter and tol (and any random_state).
        _, log_resp = self._e_step(X)

        return log_resp.argmax(dim=1)

    def _e_step(self, X):
        """E step.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)

        Returns
        -------
        log_prob_norm : float
            Mean of the logarithms of the probabilities of each sample in X

        log_responsibility : array, shape (n_samples, n_components)
            Logarithm of the posterior probabilities (or responsibilities) of
            the point of each sample in X.
        """
        log_prob_norm, log_resp = self._estimate_log_prob_resp(X)
        return torch.mean(log_prob_norm), log_resp

    def score_samples(self, X):
        """Compute the log-likelihood of each sample.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.

        Returns
        -------
        log_prob : array, shape (n_samples,)
            Log-likelihood of each sample in `X` under the current model.
        """
        check_is_fitted(self)

        return torch.logsumexp(self._estimate_weighted_log_prob(X), dim=1)

    def score(self, X, y=None):
        """Compute the per-sample average log-likelihood of the given data X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_dimensions)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        log_likelihood : float
            Log-likelihood of `X` under the Gaussian mixture model.
        """
        return self.score_samples(X).mean()

    def predict(self, X):
        """Predict the labels for the data samples in X using trained model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.

        Returns
        -------
        labels : array, shape (n_samples,)
            Component labels.
        """
        check_is_fitted(self)
        return self._estimate_weighted_log_prob(X).argmax(dim=1)

    def predict_proba(self, X):
        """Evaluate the components' density for each sample.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.

        Returns
        -------
        resp : array, shape (n_samples, n_components)
            Density of each Gaussian component for each sample in X.
        """
        check_is_fitted(self)
        _, log_resp = self._estimate_log_prob_resp(X)
        return torch.exp(log_resp)

    def sample(self, n_samples=1):
        """Generate random samples from the fitted Gaussian distribution.

        Parameters
        ----------
        n_samples : int, default=1
            Number of samples to generate.

        Returns
        -------
        X : array, shape (n_samples, n_features)
            Randomly generated sample.

        y : array, shape (nsamples,)
            Component labels.
        """
        if n_samples < 1:
            raise ValueError(
                "Invalid value for 'n_samples': %d . The sampling requires at "
                "least one sample." % (self.n_components)
            )

        _, n_features = self.means_.shape
        n_samples_comp = np.random.multinomial(n_samples, self.weights_.cpu().numpy()).tolist()

        if self.covariance_type == "full":
            X = torch.vstack(
                [
                    MultivariateNormal(loc=mean, covariance_matrix=covariance.to(mean.device)).sample((sample,))
                    for (mean, covariance, sample) in zip(self.means_, self.covariances_, n_samples_comp)
                ]
            )
        elif self.covariance_type == "tied":
            X = torch.vstack(
                [
                    MultivariateNormal(loc=mean, covariance_matrix=self.covariances_).sample((sample,))
                    for (mean, sample) in zip(self.means_, n_samples_comp)
                ]
            )
        else:
            X = torch.vstack(
                [
                    mean + torch.randn(size=(sample, n_features), device=covariance.device) * torch.sqrt(covariance)
                    for (mean, covariance, sample) in zip(self.means_, self.covariances_, n_samples_comp)
                ]
            )

        y = torch.concatenate(
            [torch.ones(sample, dtype=torch.int, device=X.device) * j for j, sample in enumerate(n_samples_comp)]
        )

        return (X, y)

    def _estimate_weighted_log_prob(self, X):
        """Estimate the weighted log-probabilities, log P(X | Z) + log weights.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)

        Returns
        -------
        weighted_log_prob : array, shape (n_samples, n_component)
        """
        return self._estimate_log_prob(X) + self._estimate_log_weights().to(X.device)

    def _estimate_log_prob_resp(self, X):
        """Estimate log probabilities and responsibilities for each sample.

        Compute the log probabilities, weighted log probabilities per
        component and responsibilities for each sample in X with respect to
        the current state of the model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)

        Returns
        -------
        log_prob_norm : array, shape (n_samples,)
            log p(X)

        log_responsibilities : array, shape (n_samples, n_components)
            logarithm of the responsibilities
        """
        weighted_log_prob = self._estimate_weighted_log_prob(X)
        log_prob_norm = torch.logsumexp(weighted_log_prob, dim=1)
        log_resp = weighted_log_prob - log_prob_norm[:, np.newaxis]
        return log_prob_norm, log_resp


def test():
    import itertools

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    n = 100
    d = 2
    c = 3

    covariance_types = ["full", "tied", "diag"]
    init_params_types = ["random", "random_from_data"]
    for cov_type, init_param in itertools.product(covariance_types, init_params_types):
        print(f"Testing {cov_type} {init_param}...")
        gmm = GaussianMixture(n_components=c, covariance_type=cov_type, max_iter=100, tol=1e-3, init_params=init_param)
        X = torch.randn(n, d, dtype=torch.float32).to(device)
        gmm.fit(X)
        params = gmm._get_parameters()
        for p in params:
            print(p.device, p.shape, p.dtype)
        gmm.predict(X)
        gmm.predict_proba(X)
        gmm.score(X)
        scores = gmm.score_samples(X)
        gmm.sample(10)

        assert gmm.means_.shape == (c, d)
        assert scores.shape == (100,)


if __name__ == "__main__":
    test()



from typing import Callable, Optional

from torchvision.datasets import CIFAR10, CIFAR100
from torchvision.datasets.utils import verify_str_arg


class CIFAR10Wrapped(CIFAR10):
    splits = ("train", "test")

    def __init__(
        self, root: str, split: str = "test", transform: Optional[Callable] = None, download: bool = False, **kwargs
    ) -> None:
        self.split = verify_str_arg(split, "split", self.splits)

        super().__init__(root, train=split == "train", transform=transform, download=download, **kwargs)


class CIFAR100Wrapped(CIFAR100):
    splits = ("train", "test")

    def __init__(
        self, root: str, split: str = "test", transform: Optional[Callable] = None, download: bool = False, **kwargs
    ) -> None:
        self.split = verify_str_arg(split, "split", self.splits)

        super().__init__(root, train=split == "train", transform=transform, download=download, **kwargs)


CIFAR10_LABELS = (
    "airplane",
    "automobile",
    "bird",
    "cat",
    "deer",
    "dog",
    "frog",
    "horse",
    "ship",
    "truck",
)

CIFAR100_CATS = {
    "aquatic mammals": ["beaver", "dolphin", "otter", "seal", "whale"],
    "fish": ["aquarium fish", "flatfish", "ray", "shark", "trout"],
    "flowers": ["orchids", "poppies", "roses", "sunflowers", "tulips"],
    "food containers": ["bottles", "bowls", "cans", "cups", "plates"],
    "fruit and vegetables": ["apples", "mushrooms", "oranges", "pears", "sweet", "peppers"],
    "household electrical devices": ["clock", "computer", "keyboard", "lamp", "telephone", "television"],
    "household furniture": ["bed", "chair", "couch", "table", "wardrobe"],
    "insects": ["bee", "beetle", "butterfly", "caterpillar", "cockroach"],
    "large carnivores": ["bear", "leopard", "lion", "tiger", "wolf"],
    "large man-made outdoor things": ["bridge", "castle", "house", "road", "skyscraper"],
    "large natural outdoor scenes": ["cloud", "forest", "mountain", "plain", "sea"],
    "large omnivores and herbivores": ["camel", "cattle", "chimpanzee", "elephant", "kangaroo"],
    "medium-sized mammals": ["fox", "porcupine", "possum", "raccoon", "skunk"],
    "non-insect invertebrates": ["crab", "lobster", "snail", "spider", "worm"],
    "people": ["baby", "boy", "girl", "man", "woman"],
    "reptiles": ["crocodile", "dinosaur", "lizard", "snake", "turtle"],
    "small mammals": ["hamster", "mouse", "rabbit", "shrew", "squirrel"],
    "trees": ["maple", "oak", "palm", "pine", "willow"],
    "vehicles 1": ["bicycle", "bus", "motorcycle", "pickup" "truck", "train"],
    "vehicles 2": ["lawn-mower", "rocket", "streetcar", "tank", "tractor"],
}

CIFAR100_LABELS = (
    "apple",
    "aquarium_fish",
    "baby",
    "bear",
    "beaver",
    "bed",
    "bee",
    "beetle",
    "bicycle",
    "bottle",
    "bowl",
    "boy",
    "bridge",
    "bus",
    "butterfly",
    "camel",
    "can",
    "castle",
    "caterpillar",
    "cattle",
    "chair",
    "chimpanzee",
    "clock",
    "cloud",
    "cockroach",
    "couch",
    "cra",
    "crocodile",
    "cup",
    "dinosaur",
    "dolphin",
    "elephant",
    "flatfish",
    "forest",
    "fox",
    "girl",
    "hamster",
    "house",
    "kangaroo",
    "keyboard",
    "lamp",
    "lawn_mower",
    "leopard",
    "lion",
    "lizard",
    "lobster",
    "man",
    "maple_tree",
    "motorcycle",
    "mountain",
    "mouse",
    "mushroom",
    "oak_tree",
    "orange",
    "orchid",
    "otter",
    "palm_tree",
    "pear",
    "pickup_truck",
    "pine_tree",
    "plain",
    "plate",
    "poppy",
    "porcupine",
    "possum",
    "rabbit",
    "raccoon",
    "ray",
    "road",
    "rocket",
    "rose",
    "sea",
    "seal",
    "shark",
    "shrew",
    "skunk",
    "skyscraper",
    "snail",
    "snake",
    "spider",
    "squirrel",
    "streetcar",
    "sunflower",
    "sweet_pepper",
    "table",
    "tank",
    "telephone",
    "television",
    "tiger",
    "tractor",
    "train",
    "trout",
    "tulip",
    "turtle",
    "wardrobe",
    "whale",
    "willow_tree",
    "wolf",
    "woman",
    "worm",
)



import torch
from sklearn.preprocessing import QuantileTransformer
from torch import Tensor


class QuantileAggregation:
    def fit(self, stack: Tensor, *args, **kwargs):
        stack = stack.detach().cpu().numpy()
        self.method = QuantileTransformer(n_quantiles=100, output_distribution="uniform")
        self.method.fit(stack)

    def __call__(self, scores: Tensor, *args, **kwargs):
        device = scores.device
        scores = scores.detach().cpu().numpy()
        return torch.from_numpy(self.method.transform(scores).mean(1)).to(device).view(-1)



import logging
import types
from functools import partial


from sklearn.preprocessing import (
    MaxAbsScaler,
    MinMaxScaler,
    Normalizer,
    PowerTransformer,
    QuantileTransformer,
    RobustScaler,
    StandardScaler,
)
from torch import Tensor

from .basics import none_preprocessing

_logger = logging.getLogger(__name__)

preprocessings_registry = {
    "none": none_preprocessing,
    "minmax": MinMaxScaler,
    "standard": StandardScaler,
    "robust": RobustScaler,
    "maxabs": MaxAbsScaler,
    "quantile_uniform": QuantileTransformer,
    "normalizer_l2": Normalizer,
    "power": PowerTransformer,
}


class Preprocessing:
    """Aggregation wrapper class."""

    def __init__(self, preprocessing_method, **kwargs) -> None:
        self.method = preprocessing_method

    def fit(self, x: Tensor, *args, **kwargs):
        if not hasattr(self.method, "fit"):
            _logger.debug("Aggregation method does not have a `fit` method.")
            return
        x = x.detach().cpu().numpy()
        self.method.fit(x)

    def transform(self, x: Tensor, *args, **kwargs):
        x = x.detach().cpu().numpy()
        try:
            return self.method.transform(x, *args, **kwargs)
        except Exception as e:
            return self.method(x, *args, **kwargs)

    def __call__(self, x: Tensor, *args, **kwargs):
        return self.transform(x, *args, **kwargs)


def register_preprocessing(name: str):
    """Decorator to register a new preprocessing method."""

    def decorator(f):
        preprocessings_registry[name] = f
        return f

    return decorator


def create_preprocessing(preprocessing_name: str, **kwargs) -> Preprocessing:
    if preprocessing_name not in preprocessings_registry:
        raise ValueError(f"Unknown aggregation method: {preprocessing_name}")
    if not isinstance(preprocessings_registry[preprocessing_name], types.FunctionType):
        return Preprocessing(preprocessings_registry[preprocessing_name](**kwargs))
    return Preprocessing(partial(preprocessings_registry[preprocessing_name], **kwargs))


def list_preprocessings() -> list:
    return list(preprocessings_registry.keys())



"""
Generalized detection methods templates.
"""
import logging
from abc import ABC, abstractmethod
from typing import Dict, List, Optional

import torch
import torch.distributed
from torch import Tensor, nn
from torchvision.models.feature_extraction import create_feature_extractor
from tqdm import tqdm

from detectors.aggregations import create_aggregation
from detectors.utils import sync_tensor_across_gpus

from .utils import add_output_op, create_reduction

_logger = logging.getLogger(__name__)


class Detector(ABC):
    """Detector base class."""

    def __init__(self, **kwargs):
        pass

    def start(self, example: Optional[Tensor] = None, fit_length: Optional[int] = None, *args, **kwargs):
        """Setup detector for fitting parameters.

        Args:
            example (Optional[Tensor], optional): Input example. Useful for pre-allocating memory.
                Defaults to None.
            fit_length (Optional[int], optional): Length of the fitting dataset. Useful for pre-allocating memory.
                Defaults to None.

        This is called before the first call to `update` and is optional.
        """
        pass

    def update(self, x: Tensor, y: Tensor, *args, **kwargs):
        """Accumulate features for detector.

        Args:
            x (Tensor): input tensor.
            y (Tensor): target tensor.

        This is called for each batch of the fitting dataset and is optional.
        """
        pass

    def end(self, *args, **kwargs):
        """Finalize detector fitting process.

        This is called after the last call to `update` and is optional.
        """
        pass

    def fit(self, dataloader, **kwargs):
        """Fit detector to a dataset.

        Args:
            dataloader (Dataloader): Dataloader for the fitting dataset.
        """
        fit_length = len(dataloader.dataset)
        # get example
        x, y = next(iter(dataloader))
        self.start(example=x, fit_length=fit_length, **kwargs)
        for x, y in dataloader:
            self.update(x, y, **kwargs)
        self.end(**kwargs)
        return self

    @abstractmethod
    def __call__(self, x: Tensor) -> Tensor:
        """Compute scores for each input at test time.

        Args:
            x (Tensor): input tensor.

        Returns:
            Tensor: scores for each input.
        """
        raise NotImplementedError


class DetectorWrapper(Detector):
    """Detector interface."""

    def __init__(self, detector, **kwargs):
        self.detector = detector
        if hasattr(self.detector, "model"):
            self.model = self.detector.model
            self.detector.model.eval()
        elif hasattr(self.detector, "keywords") and "model" in self.detector.keywords:
            self.model = self.detector.keywords["model"]
        else:
            self.model = None
        self.keywords = kwargs
        if self.model is not None:
            self.device = next(self.model.parameters()).device
        else:
            self.device = torch.device("cpu")

    def start(self, example: Optional[Tensor] = None, fit_length: Optional[int] = None, *args, **kwargs):
        if not hasattr(self.detector, "start"):
            _logger.warning("Detector does not have a start method.")
            return
        if example is not None:
            example = example.to(self.device)
        self.detector.start(example, fit_length, *args, **kwargs)

    def update(self, x: Tensor, y: Tensor, *args, **kwargs):
        if not hasattr(self.detector, "update"):
            _logger.warning("Detector does not have an update method.")
            return
        x = x.to(self.device)
        y = y.to(self.device)
        self.detector.update(x, y, *args, **kwargs)

    def end(self, *args, **kwargs):
        if not hasattr(self.detector, "end"):
            _logger.warning("Detector does not have an end method.")
            return
        self.detector.end(*args, **kwargs)

    def fit(self, dataloader, **kwargs):
        # get fit length # CHECK BUG
        fit_length = len(dataloader.dataset)
        # get example
        x, y = next(iter(dataloader))
        self.start(example=x, fit_length=fit_length, **kwargs)
        for x, y in dataloader:
            self.update(x, y, **kwargs)
        self.end(**kwargs)
        return self

    def __call__(self, x: Tensor) -> Tensor:
        x = x.to(self.device)

        return self.detector(x)

    def set_hyperparameters(self, **params):
        """Set the parameters of the detector."""
        model = params.pop("model", self.model)
        self.keywords.update(params)
        if hasattr(self.detector, "keywords"):
            self.detector.keywords.update(**params)
        else:
            self.detector = self.detector.__class__(model=model, **self.keywords)
        return self

    def save_params(self, path):
        """Save the parameters of the detector."""
        raise NotImplementedError

    def load_params(self, path):
        """Load the parameters of the detector."""
        raise NotImplementedError

    def __repr__(self):
        """Return the string representation of the detector."""
        return f"{self.__class__.__name__}()"


class DetectorWithFeatureExtraction(Detector):
    """Base class for OOD detectors with feature extraction.

    Args:
        model (nn.Module): Model to be used to extract features
        features_nodes (Optional[List[str]]): List of strings that represent the feature nodes.
            Defaults to None.
        all_blocks (bool, optional): If True, use all blocks of the model. Defaults to False.
        last_layer (bool, optional): If True, use also the last layer of the model. Defaults to False.
        pooling_op_name (str, optional): Pooling operation to be applied to the features.
            Can be one of:
                `max`, `avg`, `none`, `flatten`, `getitem`, `avg_or_getitem`, `max_or_getitem`.
            Defaults to "avg".
        aggregation_method_name (str, optional): Aggregation method to be applied to the features. Defaults to None.
        **kwargs
    """

    def __init__(
        self,
        model: nn.Module,
        features_nodes: Optional[List[str]] = None,
        all_blocks: bool = False,
        last_layer: bool = False,
        pooling_op_name: str = "avg_or_getitem",
        aggregation_method_name: Optional[str] = "mean",
        **kwargs,
    ):
        self.model = model
        self.model.eval()
        self.features_nodes = features_nodes
        self.all_blocks = all_blocks
        self.pooling_op_name = pooling_op_name
        self.aggregation_method_name = aggregation_method_name or "none"

        # feature feature reduction operation
        self.reduction_op = create_reduction(self.pooling_op_name)

        if self.features_nodes is not None:
            # if features nodes were explicitly specified, use them
            pass
        elif hasattr(self.model, "feature_info") and self.all_blocks:
            # if all_blocks is True, use all blocks of the model
            self.features_nodes = [fi["module"] for fi in self.model.feature_info][1:]  # type: ignore
        else:
            # extract from the penultimate layer only
            self.features_nodes = [list(self.model._modules.keys())[-2]]

        if last_layer:
            # if last_layer is True, use the last layer of the model
            self.last_layer_name = list(self.model._modules.keys())[-1]
            if self.features_nodes is None:
                self.features_nodes = [self.last_layer_name]
            else:
                self.features_nodes.append(self.last_layer_name)
        # remove duplicates
        self.features_nodes = list(set(self.features_nodes))
        _logger.info("Using features nodes: %s", self.features_nodes)

        self.feature_extractor = create_feature_extractor(self.model, self.features_nodes)
        self.feature_extractor.eval()

        # insert reduction operation after each node
        def output_reduce(x: Dict[str, Tensor]):
            return {k: self.reduction_op(v) for k, v in x.items()}

        self.feature_extractor = add_output_op(self.feature_extractor, output_reduce)
        self.aggregation_method = create_aggregation(self.aggregation_method_name, **kwargs)

        self.train_features = {}
        self.train_targets = []
        self.idx = 0

    @torch.no_grad()
    def start(self, example: Optional[Tensor] = None, fit_length: Optional[int] = None, *args, **kwargs):
        self.train_features = {}
        self.train_targets = []
        self.idx = 0
        if example is not None and fit_length is not None:
            self.feature_extractor.to(example.device)
            example_output = self.feature_extractor(example)
            for node_name, v in example_output.items():
                _logger.debug((fit_length,) + v.shape[1:])
                self.train_features[node_name] = torch.empty((fit_length,) + v.shape[1:], dtype=v.dtype)
                self.train_targets = torch.empty((fit_length,), dtype=torch.long)

    @torch.no_grad()
    def update(self, x: Tensor, y: Tensor, *args, **kwargs):
        self.batch_size = x.shape[0]
        # self.feature_extractor.to(x.device)
        features: Dict[str, Tensor] = self.feature_extractor(x)

        for node_name, v in features.items():
            v = sync_tensor_across_gpus(v).cpu()
            if node_name not in self.train_features:
                self.train_features[node_name] = [v]
            elif isinstance(self.train_features[node_name], list):
                self.train_features[node_name].append(v)
            else:
                self.train_features[node_name][self.idx : self.idx + v.shape[0]] = v

        y = sync_tensor_across_gpus(y).cpu()
        if isinstance(self.train_targets, list):
            self.train_targets.append(y)
        else:
            self.train_targets[self.idx : self.idx + y.shape[0]] = y

        self.idx += y.shape[0]

    def end(self, *args, **kwargs):
        for node_name, v in self.train_features.items():
            if isinstance(v, list):
                self.train_features[node_name] = torch.cat(v, dim=0)
            else:
                self.train_features[node_name] = v[: self.idx]
        if isinstance(self.train_targets, list):
            self.train_targets = torch.cat(self.train_targets, dim=0)
        else:
            self.train_targets = self.train_targets[: self.idx]

        self._fit_params()

        _logger.debug("Fitting aggregator %s...", self.aggregation_method_name)
        self.batch_size = self.train_targets.shape[0]  # type: ignore
        all_scores = torch.zeros(self.train_targets.shape[0], len(self.train_features))
        for i, (k, v) in tqdm(enumerate(self.train_features.items())):
            idx = 0
            for idx in range(0, v.shape[0], self.batch_size):
                all_scores[:, i] = self._layer_score(v[idx : idx + self.batch_size], k, i).view(-1)
        self.aggregation_method.fit(all_scores, self.train_targets)

        # TODO: compile graph with _layer_score

        del self.train_features
        del self.train_targets

    @abstractmethod
    def _fit_params(self) -> None:
        """Fit the data to the parameters of the detector."""
        pass

    @abstractmethod
    def _layer_score(self, features: Tensor, layer_name: Optional[str] = None, index: Optional[int] = None, **kwargs):
        """Compute the anomaly score for a single layer.

        Args:
            features (Tensor): features input tensor.
            layer_name (str, optional): name of the layer. Defaults to None.
            index (int, optional): index of the layer in the feature extractor. Defaults to None.
        """
        raise NotImplementedError

    @torch.no_grad()
    def __call__(self, x: Tensor) -> Tensor:
        # self.feature_extractor.to(x.device)
        features = self.feature_extractor(x)
        all_scores = torch.zeros(x.shape[0], len(features), device=x.device)
        for i, (k, v) in enumerate(features.items()):
            all_scores[:, i] = self._layer_score(v, k, i).view(-1)

        all_scores = self.aggregation_method(all_scores)
        return all_scores.view(-1)


class SimpleFeatureExtractor:
    def __init__(self, model):
        self.model = model

    def __call__(self, x):
        return {"features": self.model(x)}


class DetectorWithSimpleFE(DetectorWithFeatureExtraction, ABC):
    """Detector that uses the forward pass of the model."""

    def __init__(self, model: nn.Module, pooling_op_name: str = "avg_or_getitem", **kwargs):
        self.model = model
        self.feature_extractor = SimpleFeatureExtractor(self.model)
        self.pooling_op_name = pooling_op_name

        self.reduction_op = create_reduction(self.pooling_op_name)
        self.aggregation_method = create_aggregation("none")

        self.train_features = {}
        self.train_targets = []
        self.idx = 0

    @abstractmethod
    def _fit_params(self) -> None:
        """Fit the data to the parameters of the detector."""
        pass

    @abstractmethod
    def _layer_score(self, features: Tensor, layer_name: Optional[str] = None, index: Optional[int] = None, **kwargs):
        """Compute the anomaly score for a single layer.

        Args:
            features (Tensor): features input tensor.
            layer_name (str, optional): name of the layer. Defaults to None.
            index (int, optional): index of the layer in the feature extractor. Defaults to None.
        """
        raise NotImplementedError



import logging
from typing import Callable, Optional

import accelerate
import numpy as np
import torch
import torch.utils.data

import detectors
from detectors.pipelines import register_pipeline
from detectors.pipelines.base import Pipeline
from detectors.pipelines.ood import OODBenchmarkPipeline
from detectors.utils import ConcatDatasetsDim1

_logger = logging.getLogger(__name__)


@register_pipeline("osr_cifar10")
class OSRCifar10(OODBenchmarkPipeline):
    def __init__(
        self, transform: Callable, batch_size: int, limit_fit: Optional[int] = None, seed: int = 42, **kwargs
    ) -> None:
        super().__init__(
            "cifar10", {"cifar100": "test"}, transform, batch_size, limit_fit=limit_fit, seed=seed, **kwargs
        )


@register_pipeline("osr_cifar100")
class OSRCifar100(OODBenchmarkPipeline):
    def __init__(
        self, transform: Callable, batch_size: int, limit_fit: Optional[int] = None, seed: int = 42, **kwargs
    ) -> None:
        super().__init__(
            "cifar100", {"cifar10": "test"}, transform, batch_size, limit_fit=limit_fit, seed=seed, **kwargs
        )


@register_pipeline("osr_imagenet")
class OSRImagenet(OODBenchmarkPipeline):
    def __init__(
        self, transform: Callable, batch_size: int, limit_fit: Optional[int] = None, seed: int = 42, **kwargs
    ) -> None:
        super().__init__(
            "imagenet",
            {
                "imagenet_o": None,
            },
            transform,
            batch_size,
            limit_fit=limit_fit,
            seed=seed,
            **kwargs,
        )


@register_pipeline("one_class_versus_others_cifar10")
class SingleClassCifar10(Pipeline):
    # TODO
    def __init__(
        self,
        in_dataset_name: str,
        in_dataset_split: str,
        transform: Callable,
        batch_size: int,
        num_workers: int = 4,
        pin_memory: bool = True,
        prefetch_factor: int = 2,
        limit_fit: float = 1.0,
        limit_run: float = 1.0,
        seed: int = 42,
        **kwargs,
    ) -> None:
        self.transform = transform
        self.batch_size = batch_size
        self.limit_fit = limit_fit
        self.limit_run = limit_run
        self.seed = seed
        self.kwargs = kwargs
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.prefetch_factor = prefetch_factor
        self.in_dataset_name = in_dataset_name
        self.in_dataset_split = in_dataset_split

        self.in_dataset_name = "cifar10"
        accelerate.utils.set_seed(seed)
        self.setup()

    def _setup_datasets(self):
        """Setup `in_dataset`."""
        self.in_dataset = detectors.create_dataset(
            self.in_dataset_name, transform=self.transform, split=self.in_dataset_split
        )

    def _setup_dataloaders(self):
        if self.limit_fit is None:
            self.limit_fit = 1.0
        self.limit_fit = min(int(self.limit_fit * len(self.fit_dataset)), len(self.fit_dataset))

        # random indices
        subset = np.random.choice(np.arange(len(self.fit_dataset)), self.limit_fit, replace=False).tolist()
        self.fit_dataset = torch.utils.data.Subset(self.fit_dataset, subset)
        self.fit_dataloader = torch.utils.data.DataLoader(
            self.fit_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            prefetch_factor=self.prefetch_factor,
        )

        self.test_dataset = torch.utils.data.ConcatDataset([self.in_dataset])
        test_labels = torch.utils.data.TensorDataset(
            torch.cat(
                [torch.zeros(len(self.in_dataset))]  # type: ignore
                + [torch.ones(len(d)) * (i + 1) for i, d in enumerate(self.out_datasets.values())]  # type: ignore
            ).long()
        )

        self.test_dataset = ConcatDatasetsDim1([self.test_dataset, test_labels])
        # shuffle and subsample test_dataset
        subset = np.random.choice(
            np.arange(len(self.test_dataset)), int(self.limit_run * len(self.test_dataset)), replace=False
        ).tolist()
        self.test_dataset = torch.utils.data.Subset(self.test_dataset, subset)
        self.test_dataloader = torch.utils.data.DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            prefetch_factor=self.prefetch_factor,
        )

        _logger.info(f"Using {len(self.fit_dataset)} samples for fitting.")
        _logger.info(f"Using {len(self.test_dataset)} samples for testing.")

    def setup(self):
        self._setup_datasets()
        self._setup_dataloaders()



"""
Module containing evaluation metrics.
"""
from typing import Dict

import numpy as np
import sklearn
import sklearn.metrics
import torch
from torch import Tensor


def fpr_at_fixed_tpr(fprs: np.ndarray, tprs: np.ndarray, thresholds: np.ndarray, tpr_level: float = 0.95):
    """Return the FPR at a fixed TPR level.

    Args:
        fprs (np.ndarray): False positive rates.
        tprs (np.ndarray): True positive rates.
        thresholds (np.ndarray): Thresholds.
        tpr_level (float, optional): TPR level. Defaults to 0.95.

    Returns:
        Tuple[float, float, float]: FPR, TPR, threshold.
    """
    # return np.interp(tpr_level, tprs, fprs)
    if all(tprs < tpr_level):
        raise ValueError(f"No threshold allows for TPR at least {tpr_level}.")
    idxs = [i for i, x in enumerate(tprs) if x >= tpr_level]
    idx = min(idxs)
    return float(fprs[idx]), float(tprs[idx]), float(thresholds[idx])


def compute_detection_error(fpr: float, tpr: float, pos_ratio: float):
    """Compute the detection error.

    Args:
        fpr (float): False positive rate at a fixed TPR.
        tpr (float): True positive rate.
        pos_ratio (float): Ratio of positive labels.

    Returns:
        float: Detection error.
    """
    # Get ratios of positives to negatives
    neg_ratio = 1 - pos_ratio
    # Get indexes of all TPR >= fixed tpr level
    detection_error = pos_ratio * (1 - tpr) + neg_ratio * fpr
    return detection_error


def minimum_detection_error(fprs: np.ndarray, tprs: np.ndarray, pos_ratio: float):
    """Compute the minimum detection error.

    Args:
        fprs (np.ndarray): False positive rates.
        tprs (np.ndarray): True positive rates.
        thresholds (np.ndarray): Thresholds.
        pos_ratio (float): Ratio of positive labels.

    Returns:
        Tuple[float, float, float]: FPR, TPR, threshold.
    """
    detection_errors = [compute_detection_error(fpr, tpr, pos_ratio) for fpr, tpr in zip(fprs, tprs)]
    idx = np.argmin(detection_errors)
    return detection_errors[idx]


def get_ood_results(in_scores: Tensor, ood_scores: Tensor) -> Dict[str, float]:
    """Compute OOD detection metrics.

    Args:
        in_scores (Tensor): In-distribution scores.
        ood_scores (Tensor): Out-of-distribution scores.

    Returns:
        Dict[str, float]: OOD detection metrics.
            keys: `fpr_at_0.95_tpr`, `tnr_at_0.95_tpr`, `detection_error`, `auroc`, `aupr_in`, `aupr_out`, `thr`.
    """
    if isinstance(in_scores, np.ndarray) or isinstance(in_scores, list):
        in_scores = torch.tensor(in_scores)
    if isinstance(ood_scores, np.ndarray) or isinstance(ood_scores, list):
        ood_scores = torch.tensor(ood_scores)
    in_labels = torch.ones(len(in_scores))
    ood_labels = torch.zeros(len(ood_scores))

    _test_scores = torch.cat([in_scores, ood_scores]).cpu().numpy()
    _test_labels = torch.cat([in_labels, ood_labels]).cpu().numpy()

    fprs, tprs, thrs = sklearn.metrics.roc_curve(_test_labels, _test_scores)
    precision, recall, _ = sklearn.metrics.precision_recall_curve(_test_labels, _test_scores, pos_label=1)
    precision_out, recall_out, _ = sklearn.metrics.precision_recall_curve(_test_labels, _test_scores, pos_label=0)
    fpr, tpr, thr = fpr_at_fixed_tpr(fprs, tprs, thrs, 0.95)
    auroc = sklearn.metrics.auc(fprs, tprs)
    aupr_in = sklearn.metrics.auc(recall, precision)
    aupr_out = sklearn.metrics.auc(recall_out, precision_out)

    pos_ratio = np.mean(_test_labels == 1)
    detection_error = minimum_detection_error(fprs, tprs, pos_ratio)

    results = {
        "fpr_at_0.95_tpr": fpr,
        "tnr_at_0.95_tpr": 1 - fpr,
        "detection_error": detection_error,
        "auroc": auroc,
        "aupr_in": aupr_in,
        "aupr_out": aupr_out,
        "thr": thr,
    }
    return results



"""
Pipeline module.
"""
from enum import Enum
from typing import Any, List, Optional, Tuple

from detectors.pipelines.base import Pipeline

pipeline_registry = {}


def register_pipeline(name: str, dataset_name: Optional[str] = None):
    """
    Decorator to register a new pipeline in the registry.

    Args:
        name (str): The name of the pipeline to register.
    """

    def decorator(f):
        if dataset_name is None:
            pipeline_registry[name] = f
        else:
            pipeline_registry[name + "_" + dataset_name] = f
        return f

    return decorator


def create_pipeline(name: str, dataset_name: Optional[str] = None, **kwargs) -> Pipeline:
    """
    Utility factory method to build a Pipeline.

    Args:
        name (str, optional):
            The name defining which pipeline will be returned. Currently accepted pipeline names are:
                `ood_benchmark`, `ood_validation`, `ood_validation_noise`, `covariate_drift`, `osr`
            Complete list:
                `ood_benchmark_cifar10`, `ood_benchmark_cifar100`, `ood_benchmark_imagenet`, `ood_mnist_benchmark`,
                `ood_validation_cifar10`, `ood_validation_cifar100`, `ood_validation_imagenet`, `ood_validation_mnist`,
                `ood_validation_noise_cifar10`, `ood_validation_noise_cifar100`, `ood_validation_noise_imagenet`,
                `covariate_drift_cifar10`, `covariate_drift_cifar100`, `covariate_drift_imagenet`.

        **kwargs:
            Additional arguments to pass to the pipeline.

    Returns:
        Pipeline: A suitable pipeline for the task.

    Example::

        >>> import detectors
        >>> pipe = detectors.create_pipeline("ood_benchmark", "cifar10")
        >>> pipe.run(detector)
    """
    if dataset_name is None:
        return pipeline_registry[name](**kwargs)
    return pipeline_registry[name + "_" + dataset_name](**kwargs)


def list_pipelines() -> List[str]:
    """
    List all available pipelines.

    Returns:
        List[str]: A list of available pipelines.
    """
    return list(pipeline_registry.keys())


def list_pipeline_args(name: str) -> List[Tuple[str, Any]]:
    """
    List all available arguments for a given pipeline.

    Args:
        name (str): The name of the pipeline.

    Returns:
        list: A list of available arguments and default values for the pipeline.
    """
    import inspect

    signature = inspect.signature(pipeline_registry[name]).parameters
    return [(name, parameter.default) for name, parameter in signature.items()]


from .covariate_drift import *
from .ood import *

PipelinesRegistry = Enum("PipelinesRegistry", dict(zip(list_pipelines(), list_pipelines())))



from typing import Callable, Optional

from torchvision.datasets import MNIST, FashionMNIST
from torchvision.datasets.utils import verify_str_arg


class MNISTWrapped(MNIST):
    splits = ("train", "test")

    def __init__(
        self, root: str, split: str = "test", transform: Optional[Callable] = None, download: bool = False, **kwargs
    ) -> None:
        self.split = verify_str_arg(split, "split", self.splits)

        super().__init__(root, train=split == "train", transform=transform, download=download, **kwargs)


class FashionMNISTWrapped(FashionMNIST):
    splits = ("train", "test")

    def __init__(
        self, root: str, split: str = "test", transform: Optional[Callable] = None, download: bool = False, **kwargs
    ) -> None:
        self.split = verify_str_arg(split, "split", self.splits)

        super().__init__(root, train=split == "train", transform=transform, download=download, **kwargs)



import torch
from torch import Tensor


@torch.no_grad()
def max_logits(input: Tensor, model: torch.nn.Module, **kwargs) -> Tensor:
    """Max Logits OOD detector.

    Args:
        logits (Tensor): input tensor.

    Returns:
        Tensor: OOD scores for each input.

    References:
        [1] https://arxiv.org/abs/1911.11132
    """
    model.eval()
    logits = model(input)
    return torch.max(logits, dim=1)[0]



from typing import Callable, Optional, Tuple

import numpy as np
import torch
import torch.utils.data
from numpy.random import RandomState
from PIL import Image
from skimage.filters import gaussian as gblur


class CustomTensorDataset(torch.utils.data.Dataset):
    """TensorDataset with support for transformations.

    Args:
        *tensors (Tensor): tensors that have the same size of the first dimension.
        transform (callable, optional): transform to apply.
    """

    def __init__(self, *tensors, transform=None) -> None:
        assert all(len(tensors[0]) == len(tensor) for tensor in tensors), "Size mismatch between tensors"
        self.tensors = tensors
        self.transform = transform

    def __getitem__(self, index):
        x = Image.fromarray(self.tensors[0][index])
        if self.transform is not None:
            x = self.transform(x)

        return (x,) + tuple(tensor[index] for tensor in self.tensors[1:])

    def __len__(self):
        return len(self.tensors[0])


class Gaussian(CustomTensorDataset):
    """Gaussian noise dataset.

    Args:
        root (str): root directory.
        split (str, optional): not used.
        transform (callable, optional): transform to apply.
        download (bool, optional): not used.
        nb_samples (int): number of samples.
        shape (tuple[int, int, int]): shape of the samples.
        seed (int): seed for the random number generator.
    """

    def __init__(
        self,
        root: Optional[str] = None,
        split: Optional[str] = None,
        transform: Optional[Callable] = None,
        download: bool = False,
        nb_samples=10000,
        shape: Tuple[int, int, int] = (224, 224, 3),
        seed=1,
        **kwargs,
    ):
        rng = RandomState(seed)
        imgs = np.array(np.clip(rng.randn(nb_samples, *shape) + 0.5, 0, 1) * 255, dtype=np.uint8)
        labels = np.array([-1] * nb_samples)
        super().__init__(imgs, labels, transform=transform)


class Uniform(CustomTensorDataset):
    """Uniform noise dataset.

    Args:
        root (str): root directory.
        split (str, optional): not used.
        transform (callable, optional): transform to apply.
        download (bool, optional): not used.
        nb_samples (int): number of samples.
        shape (tuple[int, int, int]): shape of the samples.
        seed (int): seed for the random number generator.
    """

    def __init__(
        self,
        root: Optional[str] = None,
        split: Optional[str] = None,
        transform: Optional[Callable] = None,
        download: bool = False,
        nb_samples=10000,
        shape: Tuple[int, int, int] = (224, 224, 3),
        seed=1,
        **kwargs,
    ):
        rng = RandomState(seed)
        imgs = np.array(rng.rand(nb_samples, *shape) * 255, dtype=np.uint8)
        labels = np.array([-1] * nb_samples)
        super().__init__(imgs, labels, transform=transform)


class Blobs(CustomTensorDataset):
    def __init__(
        self,
        root: Optional[str] = None,
        split: Optional[str] = None,
        transform: Optional[Callable] = None,
        download: bool = False,
        nb_samples=10000,
        shape: Tuple[int, int, int] = (224, 224, 3),
        seed=1,
        **kwargs,
    ):
        """
        Blobs: amorphous shapes with definite edges.

        Args:
            root (str): root directory.
            split (str, optional): not used.
            transform (callable, optional): transform to apply.
            download (bool, optional): not used.
            nb_samples (int): number of samples.
            shape (tuple[int, int, int]): shape of the samples.
            seed (int): seed for the random number generator.

        Reference:
            [1] https://github.com/hendrycks/outlier-exposure
        """

        imgs = np.float32(np.random.binomial(n=1, p=0.7, size=(nb_samples, *shape)))
        for i in range(nb_samples):
            imgs[i] = gblur(imgs[i], sigma=1.5, channel_axis=-1)
            imgs[i][imgs[i] < 0.75] = 0.0

        # transform imgs in integers
        imgs = np.array(imgs * 255, dtype=np.uint8)
        labels = np.array([-1] * nb_samples)
        super().__init__(imgs, labels, transform=transform)


class Rademacher(CustomTensorDataset):
    def __init__(
        self,
        root: Optional[str] = None,
        split: Optional[str] = None,
        transform: Optional[Callable] = None,
        download: bool = False,
        nb_samples=10000,
        shape: Tuple[int, int, int] = (224, 224, 3),
        seed=1,
        **kwargs,
    ):
        """
        Blobs: amorphous shapes with definite edges.

        Args:
            root (str): root directory.
            split (str, optional): not used.
            transform (callable, optional): transform to apply.
            download (bool, optional): not used.
            nb_samples (int): number of samples.
            shape (tuple[int, int, int]): shape of the samples.
            seed (int): seed for the random number generator.

        Reference:
            [1] https://github.com/hendrycks/outlier-exposure
        """

        imgs = np.float32(np.random.binomial(n=1, p=0.5, size=(nb_samples, *shape)))

        # transform imgs in integers
        imgs = np.array(imgs * 255, dtype=np.uint8)
        labels = np.array([-1] * nb_samples)
        super().__init__(imgs, labels, transform=transform)



import torch
from torch import Tensor, nn

from detectors.utils import sync_tensor_across_gpus


def kl_divergence(p: Tensor, q: Tensor, eps=1e-6):
    return (p * torch.log(p / (q + eps))).sum(1)


def js_divergence(p: Tensor, q: Tensor):
    m = 0.5 * (p + q)
    return 0.5 * kl_divergence(p, m) + 0.5 * kl_divergence(q, m)


class KLMatching:
    """KL-Matching detector.

    Args:
        model (nn.Module): classifier.

    References:
        [1] https://arxiv.org/abs/1911.11132
    """

    def __init__(self, model: nn.Module, **kwargs):
        self.model = model
        self.model.eval()

        self.centroids = {}
        self.train_probs = []
        self.predictions = []

    def start(self, *args, **kwargs):
        self.centroids = {}
        self.train_probs = []
        self.predictions = []

    @torch.no_grad()
    def update(self, x: Tensor, *args, **kwargs):
        logits = self.model(x)
        logits = sync_tensor_across_gpus(logits).cpu()
        y_hat = torch.argmax(logits, dim=1).cpu()
        probs = torch.softmax(logits, dim=1).cpu()

        self.train_probs.append(probs)
        self.predictions.append(y_hat)

    def end(self):
        self.train_probs = torch.cat(self.train_probs, dim=0)
        self.predictions = torch.cat(self.predictions, dim=0)
        for c in torch.unique(self.predictions).detach().cpu().numpy().tolist():
            self.centroids[c] = torch.mean(self.train_probs[self.predictions == c], dim=0, keepdim=True)

    @torch.no_grad()
    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        probs = torch.softmax(self.model(x), dim=1)
        predictions = probs.argmax(dim=1)
        scores = torch.empty_like(predictions, dtype=torch.float32)
        for label in torch.unique(predictions).detach().cpu().numpy().tolist():
            if label not in self.centroids:
                raise ValueError(f"Label {label} not found in training set.")
            centroid = self.centroids[label].to(x.device)
            scores[predictions == label] = kl_divergence(probs[predictions == label], centroid)

        return -scores



from typing import Callable, Optional

import wilds

datasets_info = {
    "iwildcam": {
        "modality": "Image",
        "splits": ["train", "val", "test", "id_val", "id_test"],
        "unlabeled_splits": ["extra_unlabeled"],
    },
    "camelyon17": {
        "modality": "Image",
        "splits": ["train", "val", "test", "id_val"],
        "unlabeled_splits": ["train_unlabeled", "val_unlabeled", "test_unlabeled"],
    },
    "rxrx1": {
        "modality": "Image",
        "splits": ["train", "val", "test", "id_test"],
        "unlabeled_splits": [],
    },
    "ogb-molpcba": {
        "modality": "Graph",
        "splits": ["train", "val", "test"],
        "unlabeled_splits": ["train_unlabeled", "val_unlabeled", "test_unlabeled"],
    },
    "globalwheat": {
        "modality": "Image",
        "splits": ["train", "val", "test", "id_val", "id_test"],
        "unlabeled_splits": ["train_unlabeled", "val_unlabeled", "test_unlabeled", "extra_unlabeled"],
    },
    "civilcomments": {
        "modality": "Text",
        "splits": ["train", "val", "test"],
        "unlabeled_splits": ["extra_unlabeled"],
    },
    "fmow": {
        "modality": "Image",
        "splits": ["train", "val", "test", "id_val", "id_test"],
        "unlabeled_splits": ["train_unlabeled", "val_unlabeled", "test_unlabeled"],
    },
    "poverty": {
        "modality": "Image",
        "splits": ["train", "val", "test", "id_val", "id_test"],
        "unlabeled_splits": ["train_unlabeled", "val_unlabeled", "test_unlabeled"],
    },
    "amazon": {
        "modality": "Text",
        "splits": ["train", "val", "test", "id_val", "id_test"],
        "unlabeled_splits": ["val_unlabeled", "test_unlabeled", "extra_unlabeled"],
    },
    "py150": {
        "modality": "Text",
        "splits": ["train", "val", "test", "id_val", "id_test"],
        "unlabeled_splits": [],
    },
}


def make_wilds_dataset(
    dataset_name, root, split="train", transform: Optional[Callable] = None, download=False, **kwargs
):

    dataset = wilds.get_dataset(dataset_name, root_dir=root, download=download)
    assert dataset is not None
    dataset = dataset.get_subset(split, transform=transform)

    return dataset


if __name__ == "__main__":
    import detectors.config as cfg

    for ds_name in datasets_info.keys():
        print(f"Creating {ds_name} dataset")
        ds = make_wilds_dataset(ds_name, root=cfg.DATA_DIR, split="train", download=True)



import logging
from functools import partial
from typing import Callable, List, Optional

import torch
import torch.fx as fx
import torch.utils.data
from torch import Tensor
from torchvision.models.feature_extraction import create_feature_extractor

_logger = logging.getLogger(__name__)

HYPERPARAMETERS = dict(p=dict(low=0.1, high=1.0, type=float, default=0.9, step=0.05))


def reactify(m: torch.nn.Module, condition_fn: Callable, insert_fn: Callable) -> torch.nn.Module:
    graph: fx.Graph = fx.Tracer().trace(m)
    # Transformation logic here
    for node in graph.nodes:
        if condition_fn(node):
            insert_fn(node, graph)

    # Return new Module
    return fx.GraphModule(m, graph)


def condition_fn(node, equals_to: str):
    if node.name == equals_to:
        return True
    return False


def insert_fn(node, graph: fx.Graph, thr: float = 1.0):
    with graph.inserting_after(node):
        new_node = graph.call_function(torch.clip, args=(node,), kwargs={"max": thr})

        # change inputs of the next node and keep the input from the new node
        node.replace_all_uses_with(new_node)
        new_node.replace_input_with(new_node, node)


class ReAct:
    """ReAct detector.

    Args:
        model (torch.nn.Module): Model to be used to extract features
        features_nodes (Optional[List[str]]): List of strings that represent the feature nodes.
            Defaults to None.
        graph_nodes_names (Optional[List[str]]): List of strings that represent the graph nodes.
            Defaults to None.
        insert_node_fn (Callable): Function to be used to insert the node. Defaults to insert_fn.
        p (float, optional): Threshold to be used to clip the features. Defaults to 0.9.

    References:
        [1] https://arxiv.org/abs/2111.12797
    """

    LIMIT = 2_560_000

    def __init__(
        self,
        model: torch.nn.Module,
        features_nodes: Optional[List[str]] = None,
        graph_nodes_names: Optional[List[str]] = None,
        insert_node_fn: Callable = insert_fn,
        p=0.9,
        **kwargs,
    ) -> None:
        self.model = model
        self.device = next(self.model.parameters()).device
        self.model.eval()
        self.features_nodes = features_nodes
        self.graph_nodes_names = graph_nodes_names
        if self.features_nodes is None:
            self.features_nodes = [list(self.model._modules.keys())[-2]]
        self.feature_extractor = create_feature_extractor(self.model, self.features_nodes)

        self.last_layer = list(self.model._modules.values())[-1]

        self.insert_node_fn = insert_node_fn
        self.p = p

        self.thr = None
        self.training_features = {}

    def start(self, *args, **kwargs):
        self.training_features = {}

    def update(self, x: Tensor, y: Tensor) -> None:
        if len(self.training_features.keys()) > 0:
            k = list(self.training_features.keys())[0]
            if self.training_features[k].view(-1).shape[0] > self.LIMIT:
                return

        with torch.no_grad():
            features = self.feature_extractor(x)

        # accumulate training features
        if len(self.training_features) == 0:
            for k in features:
                self.training_features[k] = features[k].cpu()
        else:
            for k in features:
                self.training_features[k] = torch.cat((self.training_features[k], features[k].cpu()), dim=0)

    def end(self, *args, **kwargs):
        self.thrs = list(
            {
                k: torch.quantile(self.training_features[k].view(-1)[: self.LIMIT].to(self.device), self.p).item()
                for k in self.training_features.keys()
            }.values()
        )
        if self.graph_nodes_names is not None:
            for i, node_name in enumerate(self.graph_nodes_names):
                # add clipping node to every feature node in the graph passed in the constructor
                self.model = reactify(
                    self.model,
                    condition_fn=partial(condition_fn, equals_to=node_name),
                    insert_fn=partial(insert_fn, thr=self.thrs[i]),
                )

        _logger.info("ReAct thresholds = %s", dict(zip(self.features_nodes, self.thrs)))

        del self.training_features

    @torch.no_grad()
    def __call__(self, x: Tensor) -> Tensor:
        if self.graph_nodes_names is not None:
            logits = self.model(x)
        else:
            features = torch.clip(list(self.feature_extractor(x).values())[-1], max=self.thrs[-1])
            logits = self.last_layer(features)  # type: ignore
        return torch.logsumexp(logits, dim=-1)



"""Densenet models for CIFAR10, CIFAR100 and SVHN datasets."""
import timm
import timm.models
import torch
import torch.nn as nn
from timm.models import register_model as timm_register_model
from timm.models.densenet import _create_densenet

from detectors.data import CIFAR10_DEFAULT_MEAN, CIFAR10_DEFAULT_STD
from detectors.data.constants import CIFAR100_DEFAULT_MEAN, CIFAR100_DEFAULT_STD, SVHN_DEFAULT_MEAN, SVHN_DEFAULT_STD
from detectors.models.utils import ModelDefaultConfig, hf_hub_url_template


def _cfg(url="", **kwargs):
    num_classes = kwargs.pop("num_classes", 10)
    mean = kwargs.pop("mean", CIFAR10_DEFAULT_MEAN)
    std = kwargs.pop("std", CIFAR10_DEFAULT_STD)
    return ModelDefaultConfig(
        url=url,
        num_classes=num_classes,
        input_size=(3, 32, 32),
        pool_size=(4, 4),
        crop_pct=1,
        interpolation="bilinear",
        mean=mean,
        std=std,
        first_conv="features.conv0",
        classifier="classifier",
        **kwargs,
    )


def _cfg_pretrained(url, architecture, **kwargs):
    model = timm.create_model(architecture, pretrained=False)
    base_cfg = model.default_cfg
    num_classes = kwargs.pop("num_classes", 10)
    # replace base cfg with new arguments
    base_cfg.update(kwargs)
    base_cfg["architecture"] = architecture
    base_cfg["url"] = url
    base_cfg["num_classes"] = num_classes
    return ModelDefaultConfig(**base_cfg)


default_cfgs = {
    "densenet121_cifar10": _cfg(
        url=hf_hub_url_template("densenet121_cifar10"),
        architecture="densenet121",
    ),
    "densenet121_cifar100": _cfg(
        url=hf_hub_url_template("densenet121_cifar100"),
        num_classes=100,
        mean=CIFAR100_DEFAULT_MEAN,
        std=CIFAR100_DEFAULT_STD,
        architecture="densenet121",
    ),
    "densenet121_svhn": _cfg(
        url=hf_hub_url_template("densenet121_svhn"),
        mean=SVHN_DEFAULT_MEAN,
        std=SVHN_DEFAULT_STD,
        architecture="densenet121",
    ),
    "densenet121_camelyon17": _cfg_pretrained(
        url=hf_hub_url_template("densenet121_camelyon17"),
        architecture="densenet121",
        num_classes=2,
        input_size=(3, 96, 96),
    ),
}


def _create_densenet_small(variant, block_config, pretrained=False, **kwargs):
    default_cfg = default_cfgs[variant]

    # load timm model
    architecture = default_cfg.architecture
    model = _create_densenet(architecture, growth_rate=12, block_config=block_config, pretrained=False, **kwargs)

    # override timm config
    model.default_cfg = default_cfg
    model.pretrained_cfg = default_cfg

    # override model
    model.features.conv0 = nn.Conv2d(3, 24, kernel_size=3, padding=1, bias=False)
    model.features.norm0 = nn.Identity()
    model.features.pool0 = nn.Identity()
    model.classifier = nn.Linear(model.classifier.in_features, model.default_cfg.num_classes)

    # load weights
    if pretrained:
        model.load_state_dict(
            torch.hub.load_state_dict_from_url(model.default_cfg.url, map_location="cpu", file_name=f"{variant}.pth")
        )

    return model


def _create_densenet_pretrained(variant, pretrained=False, **kwargs):
    default_cfg = default_cfgs[variant]

    # load timm model
    model = timm.create_model(default_cfg["architecture"], pretrained=not pretrained, **kwargs)
    model.default_cfg = default_cfg
    model.classifier = nn.Linear(model.classifier.in_features, model.default_cfg["num_classes"])

    if pretrained:
        model.load_state_dict(
            torch.hub.load_state_dict_from_url(default_cfg["url"], map_location="cpu", file_name=f"{variant}.pth")
        )

    return model


@timm_register_model
def densenet121_cifar10(pretrained=False, **kwargs):
    return _create_densenet_small("densenet121_cifar10", block_config=(6, 12, 24, 16), pretrained=pretrained, **kwargs)


@timm_register_model
def densenet121_cifar100(pretrained=False, **kwargs):
    return _create_densenet_small("densenet121_cifar100", block_config=(6, 12, 24, 16), pretrained=pretrained, **kwargs)


@timm_register_model
def densenet121_svhn(pretrained=False, **kwargs):
    return _create_densenet_small("densenet121_svhn", block_config=(6, 12, 24, 16), pretrained=pretrained, **kwargs)


@timm_register_model
def densenet121_camelyon17(pretrained=False, **kwargs):
    return _create_densenet_pretrained("densenet121_camelyon17", pretrained=pretrained, **kwargs)



import logging
from typing import Any, Dict, List

import accelerate
import matplotlib.pyplot as plt
import numpy as np
import sklearn
import sklearn.metrics
import torch
import torch.nn.functional as F
import torch.utils.data
from torch import Tensor
from tqdm import tqdm

import detectors
from detectors.data import create_dataset
from detectors.methods import DetectorWrapper
from detectors.pipelines import register_pipeline
from detectors.pipelines.base import Pipeline
from detectors.utils import ConcatDatasetsDim1

_logger = logging.getLogger(__name__)


class CovariateDriftPipeline(Pipeline):
    def __init__(
        self,
        dataset_name: str,
        dataset_splits: List[str],
        transform,
        corruptions: List[str],
        intensities: List[int],
        batch_size: int = 128,
        limit_fit: float = 1.0,
        warmup_size=2000,
        seed=42,
        **kwargs,
    ) -> None:
        """Covariate Drift Pipeline.

        Covariate drift event: when moving accuracy is below threshold compared to training accuracy

        Args:
            dataset_name (str): Name of the dataset.
            dataset_splits (List[str]): List of dataset splits to use in fit and in dataset, respectively.
            transform (Callable): Transform to apply to the dataset.
            corruptions (List[str]): List of corruptions to apply.
            intensities (List[int]): List of intensities to apply.
            batch_size (int, optional): Batch size. Defaults to 128.

        """
        self.accelerator = accelerate.Accelerator()
        accelerate.utils.set_seed(seed)

        _logger.info("Creating datasets...")

        fit_dataset = create_dataset(dataset_name, split=dataset_splits[0], transform=transform)
        # shuffle fit dataset
        limit_fit = limit_fit or 1
        limit_fit = min(int(limit_fit * len(fit_dataset)), len(fit_dataset))
        indices = range(len(fit_dataset))[:limit_fit]
        indices = torch.randperm(len(fit_dataset)).numpy()[:limit_fit]
        fit_dataset = torch.utils.data.Subset(fit_dataset, indices)
        in_dataset = create_dataset(dataset_name, split=dataset_splits[1], transform=transform)
        # shuffle in dataset
        indices = torch.randperm(len(in_dataset)).numpy()
        max_dataset_size = len(in_dataset) // (len(corruptions) * len(intensities) + 1)
        self.splits = torch.arange(0, len(in_dataset), max_dataset_size)
        in_dataset = torch.utils.data.Subset(in_dataset, indices[np.arange(0, self.splits[1].item())])
        warmup_dataset = torch.utils.data.Subset(fit_dataset, range(len(fit_dataset))[:warmup_size])
        in_dataset = torch.utils.data.ConcatDataset([warmup_dataset, in_dataset])
        out_datasets = {}
        for i, corruption in enumerate(corruptions):
            out_datasets[corruption] = []
            for j, intensity in enumerate(intensities):
                _indices = indices[torch.arange(self.splits[i + j + 1].item(), self.splits[i + j + 2].item())]
                out_datasets[corruption].append(
                    torch.utils.data.Subset(
                        create_dataset(dataset_name + "_c", split=corruption, intensity=intensity, transform=transform),
                        _indices,
                    )
                )
        self.splits = self.splits.numpy() + warmup_size
        self.splits = [0] + self.splits.tolist()
        _logger.info("Data splits are: %s", self.splits)

        # increasing intensity concat dataset
        out_dataset = []
        for i, intensity in enumerate(intensities):
            # shuffle corruptions?
            out_dataset.append(
                torch.utils.data.ConcatDataset([out_datasets[corruption][i] for corruption in corruptions])
            )
        out_dataset = torch.utils.data.ConcatDataset(out_dataset)

        _logger.debug("Fit dataset size: %s", {len(fit_dataset)})
        _logger.debug("In dataset size: %s", {len(in_dataset)})
        _logger.debug("Out dataset size: %s", {len(out_dataset)})

        self.in_dataset = in_dataset
        self.out_dataset = out_dataset
        self.fit_dataset = fit_dataset
        self.batch_size = batch_size
        self.warmup_size = warmup_size
        self.limit_fit = limit_fit

        self.setup()

    def setup(self):
        test_dataset = torch.utils.data.ConcatDataset([self.in_dataset, self.out_dataset])
        test_labels = torch.utils.data.TensorDataset(
            torch.cat([torch.zeros(len(self.in_dataset))] + [torch.ones(len(self.out_dataset))]).long()  # type: ignore
        )

        self.test_dataset = ConcatDatasetsDim1([test_dataset, test_labels])
        self.test_dataloader = torch.utils.data.DataLoader(self.test_dataset, batch_size=self.batch_size, shuffle=False)
        self.fit_dataloader = torch.utils.data.DataLoader(self.fit_dataset, batch_size=self.batch_size, shuffle=True)
        self.fit_dataloader = self.accelerator.prepare(self.fit_dataloader)  # careful with this with multiple gpus
        self.test_dataloader = self.accelerator.prepare(self.test_dataloader)

    def preprocess(self, method: DetectorWrapper) -> DetectorWrapper:
        if not hasattr(method.detector, "update"):
            return method

        if self.fit_dataset is None:
            _logger.warning("Fit is not set or not supported. Returning.")
            return method

        if method.model is not None:
            method.detector.model = self.accelerator.prepare(method.detector.model)

        progress_bar = tqdm(
            range(len(self.fit_dataloader)), desc="Fitting", disable=not self.accelerator.is_local_main_process
        )
        method.start()
        for x, y in self.fit_dataloader:
            method.update(x, y)
            progress_bar.update(1)
        progress_bar.close()
        self.accelerator.wait_for_everyone()
        method.end()
        return method

    def run(self, method, model, **kwargs):
        method = method

        test_labels = torch.empty(len(self.test_dataset), dtype=torch.long)
        test_targets = torch.empty(len(self.test_dataset), dtype=torch.long)
        test_scores = torch.empty(len(self.test_dataset), dtype=torch.float)
        test_preds = torch.empty(len(self.test_dataset), dtype=torch.long)
        idx = 0
        progress_bar = tqdm(
            range(len(self.test_dataloader)), desc="Inference", disable=not self.accelerator.is_local_main_process
        )
        for x, y, labels in self.test_dataloader:

            scores = method(x)
            with torch.no_grad():
                logits = model(x)

            labels, y, scores, logits = self.accelerator.gather_for_metrics((labels, y, scores, logits))

            test_labels[idx : idx + len(x)] = labels.detach().cpu()
            test_targets[idx : idx + len(x)] = y.detach().cpu()
            test_scores[idx : idx + len(x)] = scores.detach().cpu()
            test_preds[idx : idx + len(x)] = logits.detach().cpu().argmax(1)

            idx += len(x)

            progress_bar.update(1)
        progress_bar.close()
        self.accelerator.wait_for_everyone()

        assert len(test_labels) == len(test_targets) == len(test_scores) == len(test_preds) == idx

        _logger.info("Computing metrics...")
        acc_threshold = kwargs.get("acc_threshold", 0.90)
        stride = kwargs.get("stride", 1)
        metrics = self.postprocess(
            test_scores, test_preds, test_targets, test_labels, acc_threshold=acc_threshold, stride=stride
        )

        return {
            "method": method,
            "scores": test_scores,
            "preds": test_preds,
            "targets": test_targets,
            "labels": test_labels,
            **metrics,
        }

    def postprocess(
        self,
        test_scores: Tensor,
        test_preds: Tensor,
        test_targets: Tensor,
        test_labels: Tensor,
        stride=1,
        alpha=0.99,
        acc_threshold=0.90,
        **kwargs,
    ) -> Dict[str, Any]:
        win_size = self.batch_size

        avg_warmup = test_scores[: self.warmup_size].mean().item()
        data_padded = F.pad(test_scores.unsqueeze(0), (win_size - 1, 0), "constant", avg_warmup).squeeze(0)
        moving_average = data_padded.unfold(0, win_size, stride).mean(dim=1)

        ema = test_scores.clone()
        ema[0] = avg_warmup
        for i in range(1, len(test_scores)):
            ema[i] = alpha * ema[i - 1] + (1 - alpha) * test_scores[i]

        mistakes = (test_preds != test_targets).float()
        mistakes_padded = F.pad(mistakes.unsqueeze(0), (win_size - 1, 0), "constant", 0).squeeze(0)
        moving_accuracy = 1 - mistakes_padded.unfold(0, win_size, stride).mean(dim=1)

        # define real drift event: when moving accuracy is below threshold compared to training accuracy
        acc = moving_accuracy[self.splits[1] : self.splits[1] + (self.splits[2] - self.splits[1]) // 2].mean().item()
        ref = acc_threshold * acc
        _logger.info("Original accuracy: %s", acc)
        _logger.info("Reference accuracy to detect drift: %s", ref)

        drift_labels = (moving_accuracy < ref).float()
        corr_drift = np.corrcoef(-moving_average.numpy(), drift_labels.numpy())[0, 1]
        corr_acc = np.corrcoef(moving_average.numpy(), moving_accuracy.numpy())[0, 1]

        # check error if theres is only one label on drift_labels
        if len(np.unique(drift_labels.numpy())) == 1:
            auroc_drift = 1.0
        else:
            auroc_drift = float(sklearn.metrics.roc_auc_score(drift_labels, -test_scores))

        if len(np.unique(mistakes.numpy())) == 1:
            auroc_mistakes = 1.0
        else:
            auroc_mistakes = float(sklearn.metrics.roc_auc_score(mistakes, -test_scores))

        fprs, tprs, thresholds = sklearn.metrics.roc_curve(drift_labels, -test_scores)
        fpr_drift, _, _ = detectors.eval.fpr_at_fixed_tpr(fprs, tprs, thresholds, 0.95)

        fprs, tprs, thresholds = sklearn.metrics.roc_curve(mistakes, -test_scores)
        fpr_mistakes, _, _ = detectors.eval.fpr_at_fixed_tpr(fprs, tprs, thresholds, 0.95)

        return dict(
            drift_labels=drift_labels,
            first_drift=torch.argmax(drift_labels).item(),
            ref_accuracy=ref,
            splits=self.splits,
            corr_acc=corr_acc,
            corr_drift=corr_drift,
            auroc_drift=auroc_drift,
            auroc_mistakes=auroc_mistakes,
            fpr_drift=fpr_drift,
            fpr_mistakes=fpr_mistakes,
            moving_average=moving_average,
            ema=ema,
            moving_accuracy=moving_accuracy,
            mistakes=mistakes,
        )

    def report(self, results: Dict[str, Any], subsample=1):
        print("Results:")
        print("\tCorr. Acc:", results["corr_acc"])
        print("\tCorr. Drift:", results["corr_drift"])
        print("\tAUC Drift:", results["auroc_drift"])
        print("\tAUC Mistakes:", results["auroc_mistakes"])
        print("\tFPR Drift:", results["fpr_drift"])
        print("\tFPR Mistakes:", results["fpr_mistakes"])
        print("\tFirst Drift:", results["first_drift"])
        print("\tSplits", results["splits"])

        # plot results
        mpl_colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]
        fig, ax1 = plt.subplots(1, 1, figsize=(9, 4))
        ax2 = ax1.twinx()
        # subsample values
        test_scores = results["scores"][::subsample]
        moving_average = results["moving_average"][::subsample]
        ema = results["ema"][::subsample]
        # test_labels = results["test_labels"][::subsample]
        drift_labels = results["drift_labels"][::subsample]
        mistakes = results["mistakes"][::subsample]
        moving_accuracy = results["moving_accuracy"][::subsample]

        ax1.plot(test_scores.numpy(), alpha=0.4, c=mpl_colors[1], label="score")
        ax1.plot(moving_average.numpy(), alpha=0.8, c=mpl_colors[0], label="moving avg")
        ax1.plot(ema.numpy(), alpha=0.8, c=mpl_colors[4], linewidth=2, label="ema")
        ax1.grid()

        ax2.vlines(
            self.warmup_size // subsample,
            0,
            1,
            linestyle="--",
            color="gray",
            alpha=0.5,
            linewidth=3,
            label="begin test set",
        )
        ax2.plot(drift_labels.numpy(), alpha=0.5, c=mpl_colors[2], linestyle="--", label="drift", linewidth=3)
        ax2.scatter(range(len(mistakes)), mistakes.numpy(), alpha=0.5, marker="*", c=mpl_colors[3], label="mistakes")
        ax2.plot(moving_accuracy.numpy(), alpha=0.5, c=mpl_colors[3], label="moving accuracy", linewidth=2)
        # plot reference accuracy
        ax2.axhline(
            results["ref_accuracy"], linestyle=":", color="black", alpha=0.5, linewidth=3, label="drift accuracy ref"
        )
        ax1.set_xlabel("Sample index")
        ax1.set_ylabel("Scores")
        ax2.set_ylabel("Drift")
        ax2.legend(loc="upper right")
        ax1.legend(loc="lower left")

        plt.suptitle(
            f"Corr. Acc. {results['corr_acc']:.2f}\nFPR Mistakes {results['fpr_mistakes']:.2f}, AUC Mistakes {results['auroc_mistakes']:.2f}, AUC Drift {results['auroc_drift']:.2f}"
        )


@register_pipeline("covariate_drift_cifar10")
class OneCorruptionCovariateDriftCifar10Pipeline(CovariateDriftPipeline):
    def __init__(self, transform, corruption: str, intensities: List[int], batch_size: int = 128, **kwargs) -> None:
        super().__init__("cifar10", ["train", "test"], transform, [corruption], intensities, batch_size, **kwargs)


@register_pipeline("covariate_drift_cifar100")
class OneCorruptionCovariateDriftCifar100Pipeline(CovariateDriftPipeline):
    def __init__(self, transform, corruption: str, intensities: List[int], batch_size: int = 128, **kwargs) -> None:
        super().__init__("cifar100", ["train", "test"], transform, [corruption], intensities, batch_size, **kwargs)


@register_pipeline("covariate_drift_imagenet")
class OneCorruptionCovariateDriftImagenetPipeline(CovariateDriftPipeline):
    def __init__(self, transform, corruption: str, intensities: List[int], batch_size: int = 128, **kwargs) -> None:
        super().__init__("imagenet", ["train", "val"], transform, [corruption], intensities, batch_size, **kwargs)



import logging
from functools import partial, reduce
from typing import Callable, List, Literal

import numpy as np
import torch
from torch import Tensor, nn

_logger = logging.getLogger(__name__)


def input_pre_processing(score_fn: Callable, x: Tensor, eps: float):
    x.requires_grad_()
    assert x.requires_grad
    assert eps >= 0

    scores = score_fn(x)
    scores.mean().backward()
    x = x - eps * torch.sign(-x.grad)
    return x


# Feature reductions
def flatten(data: Tensor, **kwargs):
    return torch.flatten(data, 1)


def adaptive_avg_pool2d(data: Tensor, **kwargs):
    if len(data.shape) > 2:
        return torch.flatten(nn.AdaptiveAvgPool2d((1, 1))(data), 1)
    return data


def adaptive_max_pool2d(data: Tensor, **kwargs):
    if len(data.shape) > 2:
        return torch.flatten(nn.AdaptiveMaxPool2d((1, 1))(data), 1)
    return data


def getitem(data: Tensor, **kwargs):
    if data.dim() == 3:
        return data[:, 0].clone().contiguous()
    elif data.dim() > 3:
        raise ValueError("Data must be a 3D or 2D tensor")
    return data


def avg_or_getitem(data: Tensor, **kwargs):
    if data.dim() == 3:
        return data[:, 0].clone().contiguous()
    elif data.dim() > 3:
        return torch.flatten(nn.AdaptiveAvgPool2d((1, 1))(data), 1)
    return data


def max_or_getitem(data: Tensor, **kwargs):
    if data.dim() == 3:
        return data[:, 0].clone().contiguous()
    elif data.dim() > 3:
        return torch.flatten(nn.AdaptiveMaxPool2d((1, 1))(data), 1)
    return data


def none_reduction(data: Tensor, **kwargs):
    return data


reductions_registry = {
    "flatten": flatten,
    "avg": adaptive_avg_pool2d,
    "max": adaptive_max_pool2d,
    "getitem": getitem,
    "avg_or_getitem": avg_or_getitem,
    "max_or_getitem": max_or_getitem,
    "none": none_reduction,
}


def create_reduction(reduction: str, **kwargs):
    return partial(reductions_registry[reduction], **kwargs)


def get_penultimate_layer_name(model: nn.Module):
    return list(model._modules.keys())[-2]


def get_penultimate_layer(model: nn.Module):
    return list(model._modules.values())[-2]


def get_last_layer_name(model: nn.Module):
    return list(model._modules.keys())[-1]


def get_last_layer(model: nn.Module):
    return list(model._modules.values())[-1]


# matrix operations
def torch_reduction_matrix(sigma: Tensor, reduction_method="pseudo"):
    import torch

    if reduction_method == "cholesky":
        C = torch.linalg.cholesky(sigma)
        return torch.linalg.inv(C.T)
    elif reduction_method == "svd":
        u, s, _ = torch.linalg.svd(sigma)
        return u @ torch.diag(torch.sqrt(1 / s))
    elif reduction_method == "pseudo" or reduction_method == "pinv":
        return torch.linalg.pinv(sigma)
    elif reduction_method == "inverse" or reduction_method == "inv":
        return torch.linalg.inv(sigma)
    else:
        raise ValueError(f"Unknown reduction method {reduction_method}")


def sklearn_cov_matrix_estimarion(
    x: np.ndarray,
    method: Literal[
        "EmpiricalCovariance",
        "GraphicalLasso",
        "GraphicalLassoCV",
        "LedoitWolf",
        "MinCovDet",
        "ShrunkCovariance",
        "OAS",
    ] = "EmpiricalCovariance",
    **method_kwargs,
):
    import sklearn.covariance

    try:
        method = getattr(sklearn.covariance, method)(**method_kwargs)
    except AttributeError:
        raise ValueError(f"Unknown method {method}")

    method.fit(x)
    cov_mat = method.covariance_
    _logger.debug("Cov mat determinant %s", np.linalg.det(cov_mat))
    _logger.debug("Cov mat rank %s", np.linalg.matrix_rank(cov_mat))
    _logger.debug("Cov mat condition number %s", np.linalg.cond(cov_mat))
    _logger.debug("Cov mat norm %s", np.linalg.norm(cov_mat))
    _logger.debug("Cov mat trace %s", np.trace(cov_mat))
    _logger.debug("Cov mat eigvals %s", np.linalg.eigvalsh(cov_mat))
    return method.location_, method.covariance_, method.precision_


def get_composed_attr(model, attrs: List[str]):
    return reduce(lambda x, y: getattr(x, y), attrs, model)


def add_output_op(feature_extractor, output_op: Callable) -> nn.Module:
    last_node = [n for n in feature_extractor.graph.nodes if n.op == "output"][0]
    last_node_args = last_node.args
    feature_extractor.graph.erase_node(last_node)
    nodes = [n for n in feature_extractor.graph.nodes]
    with feature_extractor.graph.inserting_after(nodes[-1]):
        new_node = feature_extractor.graph.call_function(output_op, args=last_node_args)
    nodes = [n for n in feature_extractor.graph.nodes]
    with feature_extractor.graph.inserting_after(nodes[-1]):
        feature_extractor.graph.output(new_node)
    feature_extractor.recompile()
    return feature_extractor



import logging
import os
from os.path import join
from typing import Any, Callable, Optional, Tuple

import numpy as np
from PIL import Image
from torchvision.datasets.utils import check_integrity, download_and_extract_archive

log = logging.getLogger(__name__)


class MNISTC:
    """
    MNIST-C is MNIST with corruptions for benchmarking OOD methods.

    Split can be one of `train`, `test` and `leftovers`.

    Subsets can be one of:
        `all`, `brightness`, `canny_edges`, `dotted_line`, `fog`, `glass_blur`, `identity`, `impulse_noise`,
        `motion_blur`, `rotate`, `scale`, `shear`, `shot_noise`, `spatter`, `stripe`, `translate` and `zigzag`.
    """

    splits = ["train", "test", "leftovers"]

    subsets = [
        "brightness",
        "canny_edges",
        "dotted_line",
        "fog",
        "glass_blur",
        "identity",
        "impulse_noise",
        "motion_blur",
        "rotate",
        "scale",
        "shear",
        "shot_noise",
        "spatter",
        "stripe",
        "translate",
        "zigzag",
    ]

    base_folders = ["mnist_c", "mnist_c_leftovers"]

    urls = [
        "https://zenodo.org/record/3239543/files/mnist_c.zip",
        "https://zenodo.org/record/3239543/files/mnist_c_leftovers.zip",
    ]

    filenames = [
        "mnist_c.zip",
        "mnist_c_leftovers.zip",
    ]
    tgz_md5s = [
        "4b34b33045869ee6d424616cd3a65da3",
        "c365e9c25addd5c24454b19ac7101070",
    ]

    def __init__(
        self,
        root: str,
        corruption: str,
        split: str,
        transform: Optional[Callable] = None,
        download: bool = False,
        **kwargs,
    ) -> None:
        self.root = root
        self.transform = transform

        if corruption not in self.subsets and corruption != "all":
            raise ValueError()

        if split not in self.splits:
            raise ValueError()

        self.base_folder = join(root, self.base_folders[1] if split == "leftovers" else self.base_folders[0])
        self.url = self.urls[0] if split in ["train", "test"] else self.urls[1]
        self.filename = self.filenames[0] if split in ["train", "test"] else self.filenames[1]
        self.tgz_md5 = self.tgz_md5s[0] if split in ["train", "test"] else self.tgz_md5s[1]

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError("Dataset not found or corrupted." + " You can use download=True to download it")

        self.subset = corruption

        if split == "leftovers":
            # TODO
            pass
        if corruption == "all":
            self.data = np.concatenate(
                [np.load(join(self.base_folder, s, f"{split}_images.npy")) for s in self.subsets]
            )
            self.targets = np.concatenate(
                [np.load(join(self.base_folder, s, f"{split}_labels.npy")) for s in self.subsets]
            )
        else:
            self.data = np.load(join(self.base_folder, corruption, f"{split}_images.npy"))
            self.targets = np.load(join(self.base_folder, corruption, f"{split}_labels.npy"))

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        img = self.data[index]
        target = self.targets[index]

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img.squeeze(), "L")

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    def __len__(self) -> int:
        return len(self.data)

    def _check_exists(self) -> bool:
        return os.path.exists(os.path.join(self.root, self.base_folder))

    def _check_integrity(self) -> bool:
        return check_integrity(join(self.root, self.filename), self.tgz_md5)

    def download(self) -> None:
        if self._check_integrity() and self._check_exists():
            log.debug("Files already downloaded and verified")
            return
        download_and_extract_archive(self.url, self.root, filename=self.filename, md5=self.tgz_md5)



import torch
from sklearn.ensemble import IsolationForest
from sklearn.neighbors import LocalOutlierFactor
from torch import Tensor


class AnomalyBaseAggregation:
    def __init__(self, method_class, **kwargs) -> None:
        self.method = method_class(**kwargs)

    def fit(self, stack: Tensor, *args, **kwargs):
        stack = stack.detach().cpu().numpy()
        self.method.fit(stack)

    def __call__(self, scores: Tensor, *args, **kwargs):
        device = scores.device
        scores = scores.detach().cpu().numpy()
        return torch.from_numpy(self.method.score_samples(scores)).to(device).view(-1)


class LOFAggregation(AnomalyBaseAggregation):
    def __init__(self, n_neighbors: int = 20, p=2, **kwargs) -> None:
        super().__init__(LocalOutlierFactor, n_neighbors=n_neighbors, p=p, **kwargs)


class IFAggregation(AnomalyBaseAggregation):
    def __init__(self, n_estimators=100, **kwargs) -> None:
        super().__init__(IsolationForest, n_estimators=n_estimators, **kwargs)



import os
import subprocess
from typing import Callable, Optional

from torchvision.datasets import ImageFolder
from torchvision.datasets.utils import check_integrity, extract_archive


class LSUNResized(ImageFolder):
    """`LSUN (r) <ODIN_PAPER_URL>`_ Dataset subset.

    Args:
        root (string): Root directory of dataset where directory
            exists or will be saved to if download is set to True.
        split (string, optional): The dataset split, not used.
        transform (callable, optional): A function/transform that takes in an PIL image
            and returns a transformed version. E.g, `transforms.RandomCrop`.
        download (bool, optional): If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
        **kwargs: Additional arguments passed to :class:`~torchvision.datasets.ImageFolder`.
    """

    base_folder = "LSUN_resized"
    filename = "LSUN_resize.tar.gz"
    file_md5 = "278b7b31c8cb7e804a1465a8ce03a2dc"
    url = "https://www.dropbox.com/s/moqh2wh8696c3yl/LSUN_resize.tar.gz"

    def __init__(
        self, root: str, split=None, transform: Optional[Callable] = None, download: bool = False, **kwargs
    ) -> None:
        self.root = os.path.expanduser(root)
        self.dataset_folder = os.path.join(self.root, self.base_folder)
        self.archive = os.path.join(self.root, self.filename)

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError("Dataset not found or corrupted." + " You can use download=True to download it")
        super().__init__(self.dataset_folder, transform=transform, **kwargs)

    def _check_integrity(self) -> bool:
        root = self.root
        md5 = self.file_md5
        fpath = os.path.join(root, self.filename)
        return check_integrity(fpath, md5)

    def _check_exists(self) -> bool:
        return os.path.exists(self.dataset_folder)

    def download(self) -> None:
        if self._check_integrity() and self._check_exists():
            return
        print(f"Downloading {self.filename}...")

        subprocess.run(f"wget {self.url} -P {self.root}".split(" "), capture_output=True, text=True)
        extract_archive(self.archive, self.dataset_folder)


class LSUNCroped(LSUNResized):
    """`LSUN (c) <ODIN_PAPER_URL>`_ Dataset subset.

    Args:
        root (string): Root directory of dataset where directory
            exists or will be saved to if download is set to True.
        split (string, optional): The dataset split, not used.
        transform (callable, optional): A function/transform that takes in an PIL image
            and returns a transformed version. E.g, `transforms.RandomCrop`.
        download (bool, optional): If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
        **kwargs: Additional arguments passed to :class:`~torchvision.datasets.ImageFolder`.
    """

    base_folder = "LSUN_croped"
    filename = "LSUN.tar.gz"
    file_md5 = "458a0a0ab8e5f1cb4516d7400568e460"
    url = "https://www.dropbox.com/s/fhtsw1m3qxlwj6h/LSUN.tar.gz"



import logging

import torch
from torch import Tensor

_logger = logging.getLogger(__name__)


def mahalanobis_distance_inv_fast(x: Tensor, y: Tensor, precision: Tensor):
    """Mahalanobis distance betwee x and y with an accelerated implementation.

    Args:
        x (Tensor): first point.
        y (Tensor): second point.
        precision (Tensor): inverse of the covariance matrix.
    """
    d_squared = torch.mm(torch.mm(x - y, precision), (x - y).T).diag()
    return torch.sqrt(d_squared)


class MahalanobisAggregation:
    def __init__(self, *args, **kwargs) -> None:
        self.mu = None
        self.pinv = None

    def fit(self, stack: Tensor, *args, **kwargs):
        self.mu = stack.mean(dim=0, keepdim=True)
        self.pinv = torch.linalg.pinv(torch.cov(stack.T))

    def __call__(self, scores: Tensor, *args, **kwargs):
        return -mahalanobis_distance_inv_fast(scores, self.mu.to(scores.device), self.pinv.to(scores.device))



import torch
from torch import Tensor


@torch.no_grad()
def msp(input: Tensor, model: torch.nn.Module, **kwargs) -> Tensor:
    """Maximum Softmax Response OOD detector.

    Args:
        input (Tensor): input tensor.
        model (nn.Module): classifier.

    Returns:
        Tensor: OOD scores for each input.

    References:
        [1] https://arxiv.org/abs/1610.02136
    """
    model.eval()
    logits = model(input)
    probs = torch.softmax(logits, dim=1)
    return torch.max(probs, dim=1)[0]



"""
Detectors package.
"""

from detectors.eval import get_ood_results

from . import config, data, eval, methods, models, utils
from .aggregations import create_aggregation, list_aggregations, register_aggregation
from .data import create_dataset, get_dataset_cls, list_datasets, register_dataset
from .methods import Detector, create_detector, create_hyperparameters, list_detectors, register_detector
from .methods.utils import create_reduction
from .models import create_model, create_transform, list_models, timm_register_model
from .pipelines import create_pipeline, list_pipelines, register_pipeline
from .aggregations import create_aggregation, list_aggregations, register_aggregation



import torch
from torch import Tensor


def random_score(input: Tensor, **kwargs) -> Tensor:
    return torch.rand((input.shape[0],), device=input.device)


def always_one(input: Tensor, **kwargs) -> Tensor:
    return torch.ones((input.shape[0],), device=input.device)


def always_zero(input: Tensor, **kwargs) -> Tensor:
    return torch.zeros((input.shape[0],), device=input.device)



import torch
from torch import Tensor, nn


@torch.no_grad()
def mcdropout(x: Tensor, model: nn.Module, k: int = 5, **kwargs) -> Tensor:
    """MC Dropout

    Forward-propagates the input through the model several times with activated dropout and averages the results.

    Args:
        x (Tensor): input tensor.
        model (nn.Module): classifier.
        k (int, optional): number of forward passes. Defaults to 5.

    References:
        [1] http://proceedings.mlr.press/v48/gal16.pdf
    """
    model.eval()

    has_dropout = False
    for m in model.modules():
        if isinstance(m, (nn.Dropout, nn.Dropout1d, nn.Dropout2d, nn.Dropout3d)):
            m.train()
            has_dropout = True

    if not has_dropout:
        return torch.softmax(model(x), dim=1).max(dim=1)[0]

    results = None
    for i in range(k):
        probs = torch.softmax(model(x), dim=1)
        if results is None:
            results = probs.clone()
        results += probs
    results = results / k  # type: ignore

    return results.max(dim=1)[0]



from detectors.data.cifar_wrapper import CIFAR10_LABELS, CIFAR100_LABELS
from detectors.data.cifarc import CORRUPTIONS

CIFAR10_DEFAULT_MEAN = (0.4914, 0.4822, 0.4465)
CIFAR10_DEFAULT_STD = (0.2023, 0.1994, 0.2010)
CIFAR100_DEFAULT_MEAN = (0.5071, 0.4867, 0.4408)
CIFAR100_DEFAULT_STD = (0.2675, 0.2565, 0.2761)
SVHN_DEFAULT_MEAN = (0.4377, 0.4438, 0.4728)
SVHN_DEFAULT_STD = (0.1980, 0.2010, 0.1970)
STL10_DEFAULT_MEAN = (0.43, 0.42, 0.39)
STL10_DEFAULT_STD = (0.27, 0.26, 0.27)
MNIST_DEFAULT_MEAN = (0.173,)
MNIST_DEFAULT_STD = (0.332,)
IMAGENET_DEFAULT_MEAN = (0.485, 0.456, 0.406)
IMAGENET_DEFAULT_STD = (0.229, 0.224, 0.225)
